#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import copy

from	ttproto.core.typecheck	import *
from	ttproto.core.data	import *
from	ttproto.core		import exceptions

__all__ = [
	'ListValue',
	'ListClass',
	'OrderedListClass',
	'UnorderedListClass',
]


class ListValue (Value):
	"""A class for representing a list of data

	A list is a structured value containing zero or more elements that may
	or may not be ordered (this is decided when the list class is created).

	ListValue is an abstract class. The actual classes representing the
	PDUs are generated using the metaclass function OrderedListClass() or
	UnorderedListClass().
	"""

	## class methods
	@classmethod
	@typecheck
	def metaclass_func (cls, name, bases, classdict, content_type: is_type, is_ordered: bool):
		"""Metaclass function for generating a ListValue class.

		Parameters:

		- content_type	The type of the data that can be stored in the list.

		- is_ordered	Boolean telling if the list is ordered or not. This affects
				the behaviour of the _match() method.
		"""

		assert len (bases) == 0 # FIXME: it could be useful to have multiple inheritance in some cases ?

		result = type (name, (cls,), classdict)

		result.__cls_content_type = get_type (content_type)
		result.__cls_is_ordered   = is_ordered

		return result

	@classmethod
	def is_ordered (cls):
		"""Return true if the present list class is ordered"""
		return cls.__cls_is_ordered

	@classmethod
	def get_content_type (cls):
		"""Return the type of data this list can store"""
		return cls.__cls_content_type

	## instance methods

	@typecheck
	def __init__ (self, datas: optional(iterable) = ()):
		"""Initialise a list from a set of data"""
		Value.__init__ (self)

		self.__datas = [store_data (d, self.__cls_content_type, none_is_allowed = False) for d in datas]
		self.__parent_length = 0

	@typecheck
	#TODO: clarify this point (flattened equivalent)
	def set_parent (self, parent: this_class):
		self.__parent_length = len (parent)
		return super (ListValue, self).set_parent (parent)

	def __len__ (self):
		"""return the number of elements in the list"""
		return self.__parent_length + len (self.__datas)

	@typecheck
	def __getitem__ (self, index: either(int, type)) -> optional (is_data):
		"""Access one element in the list

		If 'index' is an int, it can be either positive or negative indexing.

		If 'index' is a type, the function returns self.find_type(index) or
		raise KeyError if not found.
		"""

		if isinstance (index, type):
			result = self.find_type (index)
			if result is None:
				raise KeyError
			return result
		else:
			assert -len (self) <= index < len (self)

			if index < 0:
				# -> convert into a positive index
				index = len (self) + index

			l = self
			while index < l.__parent_length:
				l = l.get_parent()

			return l.__datas[index - l.__parent_length]

	def find_type (self: is_flat_value, type_: is_type) -> optional(Value):
		"""Find the first element of a given type in the list.

		If no value of this type is found, the function just returns None.

		NOTE: There is no recursion.
		"""
		type_ = get_type (type_)
		for d in self.__datas:
			if isinstance (d, type_):
				return d
		# FIXME: shall we look recursively into each value ?
		return None


	def __iter__ (self):
		"""Return an iterator on the data contained in this list (without the inherited elements)"""
		#FIXME: this function is not consistent with __getitem__ (because of inheritance)
		return iter (self.__datas)

	def _repr (self):
		return "%s([%s])" % (type (self).__name__, ", ".join (repr(d) for d in self.__datas))

	@typecheck
	def _is_flat (self) -> bool:
		for v in self.__datas:
			if not is_flat_value (v):
				return False
		return True

	#TODO: clarify how inheritance is handled (this may not be the best choice)
	@typecheck
	def get_datas (self) -> iterable:
		"""Return an iterator on all the elements of the list (including the inherited elements)"""
		p = self.get_parent()
		if p:
			for d in p.get_datas():
				yield d

		for d in self.__datas:
			yield d

	@classmethod
	@typecheck
	def _flatten (cls, lst_values: list_of (this_class)):

		# FIXME: will not work if the next values in lst_values are not parents of this one
		return cls ([Data.flatten (v) for v in lst_values[0].get_datas()])


	@staticmethod
	@typecheck
	# NOTE: the content of 'values' and 'patterns' is altered by this function
	def __try_all_matches (values: list_of (is_value), patterns: list_of (is_data)):
#		assert len(values) == len(patterns)

		for i in range (0, len(values)):
			for j in range (0, len (patterns)):
				if patterns[j].match (values[i]):
					v = values.pop(i)
					p = patterns.pop(j)
					if ListValue.__try_all_matches (values, patterns):
						return True
					values.insert (i, v)
					patterns.insert (j, p)

		# if the lists are empty then it is a success
		return not (values or patterns)

	@typecheck
	def _match (self, value: this_class, mismatch_list: optional(list)) -> bool:

		if len (self) == len (value):
			result = True
		else:
			result = False

			if mismatch_list is not None:
				mismatch_list.append (LengthMismatch (value, self))
			else:
				# no need to continue
				return result

		if self.__cls_is_ordered:

			# Ordered list
			for value, pattern in zip (value.get_datas(), self.get_data()):
				if not pattern.match (value):
					result = False
					if mismatch_list is None:
						# no need to continue
						break

		else:
			# Unordered list

			# FIXME: computationally intensive -> O(n!)
			if not self.__try_all_matches (list (value.get_datas()), list(self.get_datas())):

				result = False

				if mismatch_list is not None:
					mismatch_list.append (ValueMismatch (value, self))
		return result


	def _display (self, indent, output):
		indent +=2

		output.write ("###[ %s ]###\n" % self.get_type().__name__)

		for f in self.get_datas():
			f._display (indent, output)

	@typecheck
	def _build_message (self: is_flat_value) -> is_flatvalue_binary:
		"""Default list encoder

		This function encodes the list elements and concatenate them
		from left to right.
		"""

		# TODO: we'll need to keep a reference to the original template ?
		if not self:
			# empty list
			return self, b""
		else:
			# non empty
			values, bins = zip(*[v.build_message () for v in self])

			bin_result = concatenate (bins)

			return type(self) (values), concatenate (bins)

	@classmethod
	def _decode_message (cls, bin_slice, count = None):
		"""Default list decoder

		This function decodes the list from left to right.
		"""

		values = []

		def decode_field():
			nonlocal bin_slice
			v, bin_slice = cls.get_content_type().decode_message (bin_slice)
			values.append (v)

		try:
			if count is not None:
				# decode exactly 'count' entries
				for i in range (count):
					decode_field()
			else:
				# decode until the end of the slice
				while bin_slice:
					decode_field()
		except Exception as e:
			exceptions.push_location (e, cls, str(len(values)))
			raise

		return cls (values), bin_slice

ListClass = ListValue.metaclass_func

def OrderedListClass (name, bases, classdict, content_type):
	"""Metaclass function for generating an ordered ListValue class.

	Parameters:
	- content_type	The type of the data that can be stored in the list.
	"""
	return ListClass (name, bases, classdict, content_type, True)

def UnorderedListClass (name, bases, classdict, content_type):
	"""Metaclass function for generating an unordered ListValue class.

	Parameters:
	- content_type	The type of the data that can be stored in the list.
	"""
	return ListClass (name, bases, classdict, content_type, False)


@typecheck
def _diff_list_handler (lst: ListValue):
	i=0
	for value in lst:
		yield str (i), value, None
		i+=1

DifferenceList.set_handler (ListValue, _diff_list_handler)





from ..common import *


class TD_M2M_NH_10(CoAPTestCase):
    """
---
TD_M2M_NH_10:
    cfg: M2M_CFG_01
    obj: AE creates a container resource in registrar CSE via a container Create Request.
    pre: AE has created an application resource <AE> on registrar CSE. 
    ref: 'TS-0001, clause 10.2.4.1 ; TS-0004, clause 7.3.5.2.1'
    seq:
    -   s: "AE is requested to send a POST request to CSE to create a <container>"
    -   c:
        - 'The request sent by the client contains:'
        -   - Type = 0 (CON
            - Code = 2 (POST)
            - Non-empty payload
            - Content-format option
            - oneM2M-TY "3"
            - oneM2M-FR (originator) "AE_ID"
            - oneM2M-RQI "token-string" (CRQI)
    -   v: Check if possible that the <container> resource is created in registrar CSE. 
    -   c:
        - 'Registrar CSE sends response containing:'
        -   - Code = 2.01 (Created)
            - oneM2M-RSC= 2001
            - oneM2M-RQI=CRQI
            - Location-Path
            - contentFormat option
            - non-empty payload
    
    -   v: AE indicates successful operation
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]

        .. note::
            Check the number/value of the uri query options or not?
        """
        return [CoAP(type='con', code='post')]


    def run (self):
        self.match('client', CoAP (type="con", code="post",pl=Not(b"")), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionContentFormat())), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionOneM2MFrom())), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionOneM2MTY('3'))), 'fail')
        if self.match('client', CoAP(opt=Opt(CoAPOptionOneM2MRequestIdentifier())), 'fail'):
            CMID = self.coap['mid']
            CTOK = self.coap['tok']
            OPTS = self.coap['opt']
            RI = OPTS[CoAPOptionOneM2MRequestIdentifier]
            RIVAL = RI[2]

            self.next()
            self.match('server', CoAP(code=2.01, mid=CMID, tok=CTOK, pl=Not(b'')), 'fail')
            self.match('server', CoAP(opt=Opt(CoAPOptionContentFormat())), 'fail')
            self.match('server', CoAP(opt=Opt(CoAPOptionOneM2MResponseStatusCode('2001'))), 'fail')
            self.match('server', CoAP(opt=Opt(CoAPOptionOneM2MRequestIdentifier(RIVAL))), 'fail')
            self.match('server', CoAP(opt=Opt(CoAPOptionLocationPath())), 'fail')




#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from	ttproto.core.typecheck	import *
from	ttproto.core.data	import *
from	ttproto.core.exceptions	import Error

__all__ = [
	'PrimitiveValue',
	'PrimitiveValueClass',
	'IntValue',
	'BytesValue',
	'StrValue',
	'BoolValue',
]


class PrimitiveValue (Value):
	__type_map = {} # maps:	  python primitive type -> PrimitiveValueClass

	@classmethod
	def new (cls, value):
		try:
			t = cls.__type_map[type (value)]
		except KeyError:
			raise Error ("Cannot build a PrimitiveValue from '%s' (not a known python primitive type)" % value)

		return t (value)

	def __init__ (self, value = None):
		Value.__init__ (self)

	def describe (self, desc):
		# it is useless to have 'BytesValue' or 'IntValue' as message info description
		return False

	@classmethod
	@typecheck
	def get_type_for_python_type (cls, python_type: type):
		try:
			return cls.__type_map[python_type]
		except KeyError:
			raise Error ("There is no know PrimitiveValue class to handle type '%s'" % python_type)

	@classmethod
	def __register_type (cls, python_type, pv_class):
		assert issubclass (pv_class, PrimitiveValue)
		assert issubclass (pv_class, python_type) or (pv_class.__name__, python_type == "BoolValue", bool)

		cls.__type_map[python_type] = pv_class

	def get_python_type(self):
		return self.__python_type

	def _is_flat (self):
		return True

	def _repr (self):
		result = self._python_repr()

		if self.get_parent():
			result = "%s(%s)" % (type (self).__name__, result)

		return result

	def _python_repr (self):
		return self.__python_type.__repr__ (self)

	@classmethod
	def _flatten (cls, values: list_of (this_class)):

		# ensure that all the values are equal
		v0 = values[0]
		for v in values[1:]:
			if v != v0:
				raise Error ("cannot flatten different values together " + str (values))

		return v0

	def _match (self, value: optional (is_flat_value), mismatch_list: optional (list) = None) -> bool:

		if self == value:
			return True

		if mismatch_list is not None:
			mismatch_list.append (ValueMismatch (value, self))

		return False

	def _display (self, indent, output):
		indent += 2
		print("###[ %s ]###" % type(self).__name__, file = output)
		hdr = "%sValue= " % (' ' * indent)
		print("%s%s%s" % (
			hdr,
			" "* (0 if len(hdr) >= 28 else 28-len(hdr)),
			self._python_repr(),
		), file = output)

	def _build_message (self):
		return self, bytes (self.__python_type.__str__ (self), "utf-8")


def PrimitiveValueClass (name, bases, classdict):

	assert len (bases) == 1
	assert isinstance (bases[0], type)
	assert not issubclass (bases[0], Value)

	python_type = bases[0]

	# store the primitive type
	classdict["_PrimitiveValue__python_type"] = python_type

	if python_type == bool:
		# bool cannot be subclassed --> use int instead
		bases = PrimitiveValue, int
	else:
		bases = PrimitiveValue, python_type

	new_cls = type (name, bases, classdict)

	# register the new class so that we can build an instance from a primitive value
	PrimitiveValue._PrimitiveValue__register_type (python_type, new_cls)
	return new_cls

class IntValue (int, metaclass = PrimitiveValueClass):
	pass

class BytesValue (bytes, metaclass = PrimitiveValueClass):

	def _build_message (self):
		return self, self

	@classmethod
	def _decode_message (cls, bin_slice):

		value = bin_slice.raw()
		if bin_slice.get_bit_length() % 8:
			# remove the trailing byte and leave it in the slice
			value = value[:-1]
		new_slice = bin_slice[len(value):]

		return value, new_slice

# Strings are encoded in the UTF-8 format by default
class StrValue (str, metaclass = PrimitiveValueClass):
	def _build_message (self):
		return self, bytes (self, "utf-8")

	@classmethod
	def _decode_message (cls, bin_slice):

		buff = bin_slice.raw()
		if bin_slice.get_bit_length() % 8:
			# remove the trailing byte and leave it in the slice
			buff = buff[:-1]
		new_slice = bin_slice[len(buff):]

		return str (buff, "utf-8"), new_slice

class BoolValue (bool, metaclass = PrimitiveValueClass):
	def __new__ (cls, value = False):
		return int.__new__(cls, bool (value))

	def _python_repr (self):
		return repr (bool (self))

	def _build_message (self):
		assert self in (0, 1)
		return self, ((b"\x80" if self else b"\0"), 1)

	@classmethod
	def _decode_message (cls, bin_slice):
		return bin_slice[0] & 128, bin_slice.shift_bits(1)





#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *


class TD_COAP_CORE_07 (CoAPTestcase):
    """Identifier:
TD_COAP_CORE_07
Objective:
Perform PUT transaction (NON mode)
Configuration:
CoAP_CFG_BASIC
References:
[COAP] 5.8.3, 5.2.3

Pre-test
conditions:
•	Server offers a /test resource that handles PUT

Test Sequence:
Step
Type
Description

1
Stimulus
Client is requested to send a PUT request with:
•	Type = 1(NON)
•	Code = 3(PUT)
•	An arbitrary payload
•	Content-format option

2
Check
The request sent by the client contains:
•	Type=1 and Code=3
•	Client-generated Message ID (➔ CMID)
•	Client-generated Token (➔ CTOK)
•	Uri-Path option "test"

3
Verify
Server displays the received information

4
Check
Server sends response containing:
•	Type = 1(NON)
•	Code = 68 (2.04 Changed) or 65 (2.01 Created)
•	Server-generated Message ID (➔ SMID)
•	Token = CTOK
•	Content-format option if payload non-empty
•	Empty or non-empty Payload

5
Verify
Client displays the received response
"""
    def run (self):
        self.match_coap ("client", CoAP (type="non", code="put",
                        opt = self.uri ("/test")))
        self.match_coap ("client", CoAP (
                        pl  = Not (b''),
                        opt = Opt (CoAPOptionContentFormat()),
                ), "fail")
        CTOK = self.frame.coap["tok"]

        self.next()

        self.match_coap ("server", CoAP (
                        type = "non",
                        code = Any (65, 68),
                        tok = CTOK,
                ))
        if self.match_coap ("server", CoAP(pl = Not(b"")),None):
            self.match_coap ("server", CoAP (
                        opt = Opt (CoAPOptionContentFormat()),
                ), "fail")





#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from	ttproto.core.data		import *
from	ttproto.core.typecheck		import *
from	ttproto.core.lib.inet.meta	import *
from	ttproto.core.lib.inet.basics 	import *
from	ttproto.core.lib.inet.ipv6	import *
from	ttproto.core			import port

#import	ttproto.core.lib.inet.ipv6

__all__ = [
	'SixLowpan',
	'SixLowpanMESH',
	'SixLowpanIPv6',
	'SixLowpanFRAG1',
	'SixLowpanFRAGN'
]


class SixLowpan (
	metaclass = InetPacketClass,
	fields    = [
		("Dispatch",		"dp",		Bin (UInt8),	InetType (None)),
		("Payload", 		"pl", 		Value),
	]):
	"""
	defined in [RFC4944]
				1                   2                   3
	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	   |0 1| Dispatch  |  type-specific header
	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	"""
	pass

sixlowpan_dispatch_bidict = BidictValueType (0, SixLowpan, allow_duplicates = True)
SixLowpan.get_field(0).tag._set_bidict (sixlowpan_dispatch_bidict)



class SixLowpanMESH (
	metaclass  = InetPacketClass,
	variant_of = SixLowpan,
	prune      = 0,
	fields	   = [
		("Dispatch",			"dp",	Bin (UInt2),	0b10),
		("V",				"v",	bool,		0),
		("F",				"f",	bool,	 	0),
		("HopsLeft",			"hl",	UInt4,	 	0),
		("OriginatorAddress",		"src",	Value,	 	0),
		("FinalDestinationAddress",	"dst",	Value,	 	0),
		("Payload",			"pl",	SixLowpan),
	]):
	"""
	defined in [RFC4944]:

		   Pattern    Header Type
		 +------------+-----------------------------------------------+
		 | 10  xxxxxx | MESH       - Mesh Header                      |
		 +------------+-----------------------------------------------+

			      1                   2                   3
	       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
	      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	      |1 0|V|F|HopsLft| originator address, final address
	      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	"""
	pass



class SixLowpanFRAG1 (
	metaclass  = InetPacketClass,
	variant_of = SixLowpan,
	prune      = 0,

	fields	   = [
		("Dispatch",			"dp",	Bin (UInt5),	0b11000),
		("DatagramSize",		"size",	UInt11,	 	0),
		("DatagramTag",			"tag",	Hex (UInt16), 	0),
		("Payload",			"pl",	Value),
	]):
	"""
	 defined in [RFC4944]:

		   Pattern    Header Type
		 +------------+-----------------------------------------------+
		 | 11  000xxx | FRAG1      - Fragmentation Header (first)     |
		 +------------+-----------------------------------------------+
				   1                   2                   3
	       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
	      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	      |1 1 0 0 0|    datagram_size    |         datagram_tag          |
	      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	"""
	pass



class SixLowpanFRAGN (
	metaclass  = InetPacketClass,
	variant_of = SixLowpan,
	prune      = 0,
	fields	   = [
		("Dispatch",			"dp",	Bin (UInt5),	0b11100),
		("DatagramSize",		"size",	UInt11,	 	0),
		("DatagramTag",			"tag",	Hex (UInt16), 	0),
		("DatagramOffset",		"ofs",	UInt8,	 	0),
		("Payload",			"pl",	Value),
	]):
	"""
	defined in [RFC4944]:

		   Pattern    Header Type
		 +------------+-----------------------------------------------+
		 | 11  100xxx | FRAGN      - Fragmentation Header (subsequent)|
		 +------------+-----------------------------------------------+
				   1                   2                   3
	       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
	      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	      |1 1 0 0 0|    datagram_size    |         datagram_tag          |
	      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	      |datagram_offset|
	      +-+-+-+-+-+-+-+-+
	"""
	pass



class SixLowpanBroadcast  (
	metaclass  = InetPacketClass,
	variant_of = SixLowpan,
	prune      = 0,
	fields	   = [
		("Dispatch",		"dp",	Bin (UInt8),	0b01010000),
		("SequenceNumber",	"seq",	UInt8,	 	0),
		("Payload",		"pl",	SixLowpan),
	]):
	"""
	 defined in [RFC4944]:

		   Pattern    Header Type
		 +------------+-----------------------------------------------+
		 | 01  010000 | LOWPAN_BC0 - LOWPAN_BC0 broadcast             |
		 +------------+-----------------------------------------------+

				 1
	       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
	      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	      |0|1|LOWPAN_BC0 |Sequence Number|
	      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	"""
	pass


class SixLowpanIPv6 ( #FIXME: not sure it is the right way to do it.
	metaclass  = InetPacketClass,
	variant_of = SixLowpan,
	prune      = 0,
	fields	   = [
		("Dispatch",			"dp",	Bin (UInt8), 0x41),
		("Payload",			"pl",	IPv6),
	]):
	"""
	defined in [RFC4944]:

		   Pattern    Header Type
		 +------------+-----------------------------------------------+
		 | 01  000001 | IPv6       - Uncompressed IPv6 Addresses      |
		 +------------+-----------------------------------------------+
	"""

	def describe (self, desc):
		return self.describe_payload (desc)


"""
Dispatch value bit pattern defined in [RFC4944]:

           Pattern    Header Type
         +------------+-----------------------------------------------+
         | 00  xxxxxx | NALP       - Not a LoWPAN frame               |
         | 01  000001 | IPv6       - Uncompressed IPv6 Addresses      |
         | 01  000010 | LOWPAN_HC1 - LOWPAN_HC1 compressed IPv6       |
         | 01  000011 | reserved   - Reserved for future use          |
         |   ...      | reserved   - Reserved for future use          |
         | 01  001111 | reserved   - Reserved for future use          |
         | 01  010000 | LOWPAN_BC0 - LOWPAN_BC0 broadcast             |
         | 01  010001 | reserved   - Reserved for future use          |
         |   ...      | reserved   - Reserved for future use          |
         | 01  111110 | reserved   - Reserved for future use          |
         | 01  111111 | ESC        - Additional Dispatch byte follows |
         | 10  xxxxxx | MESH       - Mesh Header                      |
         | 11  000xxx | FRAG1      - Fragmentation Header (first)     |
         | 11  001000 | reserved   - Reserved for future use          |
         |   ...      | reserved   - Reserved for future use          |
         | 11  011111 | reserved   - Reserved for future use          |
         | 11  100xxx | FRAGN      - Fragmentation Header (subsequent)|
         | 11  101000 | reserved   - Reserved for future use          |
         |   ...      | reserved   - Reserved for future use          |
         | 11  111111 | reserved   - Reserved for future use          |
         +------------+-----------------------------------------------+

                   Figure 2: Dispatch Value Bit Pattern
"""

sixlowpan_dispatch_bidict[0b01000001] = SixLowpanIPv6
#sixlowpan_dispatch_bidict[0b01000010] = SixLowpanHC1	# HC1 not supported
sixlowpan_dispatch_bidict[0b01010000] = SixLowpanBroadcast
#sixlowpan_dispatch_bidict[0b01111111] = SixLowpanHC1	# ESC not supported
for i in range(0b10000000, 0b11000000):
	sixlowpan_dispatch_bidict[i] = SixLowpanMESH
for i in range(0b11000000, 0b11001000):
	sixlowpan_dispatch_bidict[i] = SixLowpanFRAG1
for i in range(0b11100000, 0b11101000):
	sixlowpan_dispatch_bidict[i] = SixLowpanFRAGN





#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from ttproto.core.typecheck import *

class Error (Exception):
	"""Generic Error exception"""

	def __init__ (self, *msg):
		"""Initialise the exception with an error message

		The elements of *msg are converted into strings separated by
		spaces (just like print()).

		If msg is empty, the error message is replaced with the name of
		the exceptiontype: type(self).__name__
		"""

		l = ["Error:"]
		if msg:
			l.extend (msg)
		else:
			if type (self) != Error:
				l.append (type (self).__name__)
			elif not len (msg):
				l.append ("<undefined>")

		self.msg = " ".join([str(v) for v in l])

	def __str__ (self):
		"""Return the exception message"""
		return self.msg

	def __repr__ (self):
		return "%s(%s)" % (type (self).__name__, (repr(self.msg) if self.msg else ""))

class UserInterrupt (Exception):
	"""Exception thrown to notify an interruption by the user.

	Contrary to the builtin KeyboardInterrupt exception, this exception is
	not raised asynchronously. In order to allow a clean interruption.
	"""
	pass

class TerminateTestcase (Exception):
	"""Exception to be throw to terminate the execution of the testcase immediately.

	This exception is handled by the Testcase class. When caught, the execution
	stops immediately, without returning any error.
	"""
	pass

class PortNotConnected (Error):
	"""Attempt to send a message to a port which is not connected"""
	pass

# TODO: use builtin class NotImplementedError
class NotImplemented (Error):
	"""Function not implemented"""
	pass

class BadConversion (Error):
	"""Value conversion error"""
	def __init__ (self, data, type_, exc):
		desc = str(exc)[7:] if isinstance (exc, Error) else str(exc)
		Error.__init__ (self, "cannot convert %s into %s (%s)" % (type(data).__name__, type_.__name__, desc))

class DecodeError (Error):
	"""Decoding error"""
	def __init__ (self, binary, target, exc):
		location = "" if not hasattr (exc, "ttproto_location") else (" at " + str(exc.ttproto_location))
		Error.__init__ (self, "cannot decode %s into %s (%s%s)" % (
			repr(binary),
			target.__name__,
			(str (exc) or repr (exc)),
			location))
class UnknownField (Error):
	"""Attempt to access an unknown field"""
	def __init__ (self, value, field_name):
		Error.__init__ (self, "value of type %s has no field named %s" % (type (value).__name__, repr(field_name)))


class ReaderError(Error):
    """
    Exception class for when a reader can't process the file
    """
    pass

class Location:
	"""A class to help locating encoding or decoding errors.

	When decoding a complex structured values, it is useful to know which
	field caused the errors. This allows recording the path of the errenous field.
	"""
	def __init__ (self):
		self.__path = []

	@typecheck
	def push (self, type_:type, field: optional(str) = None):
		"""Append a new field in the path

		- type_		the type of the current type structured type being processed
		- field		the name/index of the field that caused the exception
		"""
		self.__path.append ((type_, field))

	def __str__ (self):
		return "".join ("{%s}%s" % (t.__name__, ("." + f) if f else "") for t, f in reversed(self.__path))

@typecheck
def push_location (e: Exception, type_: type, field: optional(str) = None):
	"""Utility function used to report the location of encoding and decoding errors

	This function shall be called in an except clause in the
	decoder/encoder of structured values to indicate what is the current
	type and field.

	It attach a Location object into the thrown exception, to report the
	location.

	Exemple:
		def _build_message (self):
			...
			for field_name, field_value in ...:
				try:
					value, bin = field_value.build_message()
				except Exception as e:
					push_location (e, type(self), field_name)
					raise
	"""

	if not hasattr (e, "ttproto_location"):
		e.ttproto_location = Location()

	e.ttproto_location.push (type_, field)




#!/usr/bin/env python3
#
#  (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import re
from ttproto.core.analyzer import TestCase, is_protocol, Node, Conversation, Capture
from ttproto.core.dissector import Frame
from ttproto.core.templates import All, Not, Any, Length
from ttproto.core.typecheck import typecheck, tuple_of, optional, anything, list_of
from ttproto.core.lib.all import *
from urllib import parse
from ttproto.core.exceptions import Error


class NoStimuliFoundForTestcase(Error):
    """
    Error raised when no stimuli was defined for the testcase
    """
    pass


class FilterError(Error):
    """
    The error thrown when an error occurs into Filter object during its run
    """
    pass


class SixlowpanTestCase(TestCase):
    """
    The test case extension representing a Sixlowpan test case
    """

    @classmethod
    @typecheck
    def get_protocol(cls) -> is_protocol:
        """
        Get the protocol corresponding to this test case. This has to be
        implemented into the protocol's common test case class.

        :return: The protocol on which this TC will occur
        :rtype: Value
        """
        return SixLowpanIPHC

    @typecheck
    def preprocess(
        self,
        capture: Capture
    ) -> (list_of(Conversation), list_of(Frame)):
        """
        Preprocess and filter the frames of the capture into test case related
        conversations.

        :param Capture: The capture which will be filtered/preprocessed
        :return:
        """

        # TODO assert is subclass of TesCase?

        # Get informations from the test case
        # TODO get attrbutes stimuli , protocol under test, nodes patterns
        #      directly from child's atrib?
        stimulis = self.get_stimulis()
        protocol = self.get_protocol()
        nodes = self.get_nodes_identification_templates()
        conversations = []
        ignored = []

        # TODO what happens if no protocol declared on the test case?
        if not nodes or len(nodes) < 2:
            raise ValueError(
                'Expected at leaset two nodes declaration from the test case'
            )
        if not protocol:
            raise ValueError(
                'Expected a protocol under test declaration from the test case'
            )
        # If there is no stimuli at all
        if not stimulis or len(stimulis) == 0:
            raise NoStimuliFoundForTestcase(
                'Expected stimuli declaration from the test case'
            )

        # Get the frames filtered on the protocol
        frames, ignored = Frame.filter_frames(capture.frames, protocol)

        # Get a counter of the current stimuli
        sti_count = 0
        current_conversation = None
        nb_stimulis = len(stimulis)
        for frame in frames:

            # If the frame matches a stimuli
            if stimulis[sti_count].match(frame[protocol]):

                # If it's the first stimuli
                if sti_count == 0:

                    # If there is already a conversation pending, save it
                    if current_conversation:
                        self._conversations.append(current_conversation)

                    # Get the nodes as a list of nodes
                    # TODO already done at begining. why isnde the iteeration?
                    # nodes = testcase.get_nodes_identification_templates()

                    # And create the new one
                    current_conversation = Conversation(nodes)

                # If intermediate stimulis, just increment the counter
                sti_count = (sti_count + 1) % nb_stimulis

            # If there is a current_conversation, put the frame into it
            if current_conversation:
                current_conversation.append(frame)

            # If no conversation pending
            else:
                ignored.append(frame)

        # At the end, if there is a current conversation pending, close it
        if current_conversation:

            # If not all stimulis were consumed
            if sti_count != 0:
                raise FilterError(
                    'Not all stimulis were consumed, %d left and next one should have been %s'
                    %
                    (
                        nb_stimulis - sti_count,
                        stimulis[sti_count]
                    )
                )

            # Close the current conversation by adding it to list
            conversations.append(current_conversation)

        return conversations, ignored



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import re

from	ttproto.core.exceptions		import Error
from	ttproto.core.data		import Value, BidictValueType
from	ttproto.core.typecheck		import *
from	ttproto.core.lib.inet.meta	import *
from	ttproto.core.lib.inet.basics	import *
from	ttproto.core.lib.ports		import socat

__all__ = [
	'EthernetAddress',
	'Ethernet',
	'EthernetPortTap',
]


class EthernetAddress (metaclass = FixedLengthBytesClass (6)):
	__reg = re.compile ("^([0-9A-F]{2})[-:.]?([0-9A-F]{2})[-:.]?([0-9A-F]{2})[-:.]?([0-9A-F]{2})[-:.]?([0-9A-F]{2})[-:.]?([0-9A-F]{2})\\Z", re.I)

	def __new__ (cls, value = None):
		if isinstance (value, str):
			mo = re.match (cls.__reg, value)
			if not mo:
				raise Error("malformed Ethernet address")

			value = bytes ([int (v, 16) for v in mo.groups()])

		return super().__new__ (cls, value)

	def __str__ (self):
		return ":".join ([format (c, "02x") for c in self])

ethernet_type_bidict = BidictValueType (0, bytes)

class Ethernet (
	metaclass = InetPacketClass,
	fields    = [
		("DestinationAddress",	"dst",	EthernetAddress,	"00:00:00:00:00:00"),
		("SourceAddress",	"src",	EthernetAddress,	"00:00:00:00:00:00"),
		("Type",		"type",	Hex (UInt16),		InetType (ethernet_type_bidict, "Payload")),
		("Payload",		"pl",	Value),
		("Trailer",		"trl",	bytes,			b""), #TODO: fill it in build_message if necessary
	],
	descriptions = {
		"Type": {
			0x0800: "IP",
			0x0806: "ARP",
			0x8100: "VLAN-tagged frame",
			0x86DD: "IPv6",
			0x880B: "PPP",
			0x8847: "MPLS",
			0x8848: "MPLS with upstream-assigned label",
			0x8863: "PPPoE Discovery Stage",
			0x8864: "PPPoE Session Stage",
			0x9000: "Loopback",
		}
	}):

	def describe (self, desc):
		desc.hw_src = self["src"]
		desc.hw_dst = self["dst"]

		# Added by 

		if not self.describe_payload (desc):
			#TODO: integrate the map into the InetFormat

			t = self.get_description ("type")
			desc.info = t if t else "Ethernet Type %04x" % self["type"]
		return True


@typecheck
def EthernetPortTap (name: optional(str) = None):
	assert not name or re.match ("[A-Za-z0-9]+$", name)

	name_add = (",tun-name=" + name) if name else ""

	#FIXME: it is mandatory to provide an IP address, what happens if it is already used ??)
	return socat.SocatMessagePort ("TUN:10.42.42.42/31,up,iff-no-pi,tun-type=tap" + name_add , Ethernet)




#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from	ttproto.core.data		import Value
from	ttproto.core.lib.inet.meta	import *
from	ttproto.core.lib.inet.basics 	import *
from	ttproto.core.lib.inet.ipv6	import *
from	ttproto.core.lib.inet.icmpv6	import *

#import	ttproto.core.lib.inet.ipv6


__all__ = [
	'SixLowpanAddressRegistrationOption',
	'SixLowpanContextOption',
	'SixLowpanAuthoritativeBorderRouterOption',
	'SixARO',
	'SixABRO',
	'SixCO',
	'SixPI',
]

class _NDLifetimeDescription:
	"""Description class for 6lowpan-nd lifetimes (in unit of 60s)"""
	def __getitem__ (self, item):
		if not isinstance (item, int):
			return None

		mins = item % 60
		item //= 60
		hours = item % 24
		item //= 24
		days = item

		return "%dd %dh %dm" % (days, hours, mins)

"""
defined in [I-D.6lowpan-nd-15]:

   0                   1                   2                   3
   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     Type      |   Length = 2  |    Status     |   Reserved    |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |           Reserved            |     Registration Lifetime     |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                                                               |
   +                            EUI-64                             +
   |                                                               |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
"""
class SixLowpanAddressRegistrationOption (
	metaclass  = InetPacketClass,
	variant_of = ICMPv6Option,
	id         = 31, #FIXME (TBD)
	prune      = -1,
	fields	   = [
		("Status",			"st",	UInt8,	 	0),
		("Reserved",			"rsv",	Hex (UInt24), 	0),
		("RegistrationLifetime",	"lt",	UInt16,	 	0),
		("EUI64",			"eui",	Eui64Address, 	"00:00:00:00:00:00:00:00"),
	],
	descriptions = {
		"Status": {
			0:	"Success",
			1:	"Duplicate Address",
			2:	"Neighbor Cache Full",
		},
		"RegistrationLifetime": _NDLifetimeDescription()
	}):
	pass

"""
defined in [I-D.6lowpan-nd-15]:

    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     Type      |     Length    |Context Length | Res |C|  CID  |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |            Reserved           |         Valid Lifetime        |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   .                                                               .
   .                       Context Prefix                          .
   .                                                               .
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
"""
class SixLowpanContextOption (
	metaclass  = InetPacketClass,
	variant_of = ICMPv6Option,
	id         = 32, #FIXME (TBD2)
	prune      = -1,

	fields	   = [
		("ContextLength", 	"clen",	UInt8,		64),
		("Reserved1", 		"rsv1",	Bin (UInt3),	0),
		("C",			"c",	bool,		0),
		("CID",			"cid",	UInt4,		0),
		("Reserved2",		"rsv2",	Hex (UInt16),	0),
		("ValidLifetime",	"lt",	UInt16,		0),
		("ContextPrefix",	"pf",	IPv6Address,	InetPaddedPrefix(64, "clen")),
	],
	descriptions = {
		"ValidLifetime": _NDLifetimeDescription()
	}):
	pass

"""
defined in [I-D.6lowpan-nd-15]:

    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     Type      |  Length = 3   |        Version Number         |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                            Reserved                           |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                                                               |
   +                                                               +
   |                                                               |
   +                          6LBR Address                         +
   |                                                               |
   +                                                               +
   |                                                               |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
"""
class SixLowpanAuthoritativeBorderRouterOption (
	metaclass  = InetPacketClass,
	variant_of = ICMPv6Option,
	id         = 33, #FIXME (TBD3)
	prune      = -1,
	fields	   = [
		("VersionNumber",	"ver",	UInt16,		0),
		("Reserved",		"rsv",	Hex (UInt32),	0),
		("6LBRAddress",		"lbr",	IPv6Address,	"::"),
	]):
	pass

def SixLowpanPI (l=0, **kw):
	return ICMPv6PI (l=l, **kw)

#Aliases
SixARO = SixLowpanAddressRegistrationOption
SixABRO = SixLowpanAuthoritativeBorderRouterOption
SixCO = SixLowpanContextOption
SixPI = SixLowpanPI



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""Logging module"""

import sys, threading, time

from ttproto.core.typecheck	import *
from ttproto.core.exceptions	import Error
from ttproto.core		import clock

__all__ = [
	'LogEvent',
	'LogEventClass',
	'Logger',
	'ConsoleLogger',
	'LoggerGroup',
	'LoggedObject',
	'EventText',
	'EventStep',
	'log_event',
	'log',
	'step',
]

class LogEvent:
	def __init__ (self, *k):
		if len(k) != len(self.__types):
			raise Error("Expected %d parameters instead of %d" % (len(self.__types), len(k)))

		for i in range (0, len (k)):
			t = self.__types[i]

			if k[i] is None and self.__optional[i]:
				continue

			if not isinstance (k[i], t):
				raise Error("Param id %d is an instance of %s instead of %s" % (i, type(k[i]), t))

		self.__values = k

		self.__timestamp = clock.Clock.get_instance().time()

	def __getattr__ (self, name):
		assert name in self.__names

		return self.__values[self.__names.index (name)]

	@typecheck
	def __getitem__ (self, i: int):
		assert 0 <= i < len (self.__values)

		return self.__values[i]

	def get_timestamp (self):
		return self.__timestamp

	def __str__ (self):
		assert type(self).__name__[0:5] == "Event"

		return "%s %s" % (type(self).__name__[5:], " ".join ([str(v) for v in self.__values]))

	# may be reimplemented
	summary = __str__


class LogEventClass (type):

	def __new__ (cls, name, bases, dct):
		assert bases == ()
		assert "fields" in dct
		assert name[0:5] == "Event"	# just a naming convention
		bases = LogEvent,

		fields = dct.pop("fields")
		assert type (fields) == tuple
		for f in fields:
			assert 2 <= len (f) <= 3
			assert type(f) == tuple
			assert type(f[0]) == str
			if len(f) > 2:
				assert type (f[2]) == bool
			# TODO: some assertion on f[1]'s type ?

		dct["_LogEvent__names"], dct["_LogEvent__types"] = zip (*fields)
		dct["_LogEvent__optional"] = tuple (False if len(f) < 3 else f[2] for f in fields)

		return type.__new__ (cls, name, bases, dct)


#TODO: add timestamps
class Logger:
	def log_event (self, event):
		pass

	@classmethod
	@typecheck
	def set_default (cls, logger: this_class):
		cls.__default_logger = logger

	@classmethod
	def get_default (cls):
		return cls.__default_logger

	def __enter__ (self):
		return self

	def __exit__ (self, a, b, c):
		pass


class ConsoleLogger (Logger):

	__linebreak = set (("EventTestcaseStarted", "EventTestcaseTerminated", "EventSessionStarted", "EventSessionTerminated", "EventSessionAborted"))

	def __init__ (self, file = sys.stdout, summary = True):
		self.__lock = threading.Lock()
		self.__file = file
		self.__summary = summary
		self.__origin = None

	@typecheck
	def log_event (self, event: LogEvent):
		with self.__lock:
			if self.__origin is None:
				self.__origin = event.get_timestamp() // 1000 * 1000
				print ("clock origin: %d (%s)" % (self.__origin, time.ctime (self.__origin)))
			br = "\n" if type(event).__name__ in self.__linebreak else ""
			print ("%s[%03.4f] %s\n%s" % (br, (event.get_timestamp() - self.__origin), event.summary() if self.__summary else event, br), file = self.__file, end="")

			if type(event).__name__ == "EventTestcaseRuntimeError":
				for line in event.traceback.split("\n"):
					print ("             " + line, file = self.__file)


Logger.set_default (ConsoleLogger())


class LoggerGroup (Logger):
	@typecheck
	def __init__ (self, loggers: list_of (Logger) = []):
		self.__loggers = list (loggers)

	@typecheck
	def add_logger (self, logger: Logger):
		self.__loggers.push (logger)

	def log_event (self, event):
		for l in self.__loggers:
			l.log_event (event)

	def __enter__ (self):
		for l in self.__loggers:
			l.__enter__()
		return self

	def __exit__ (self, a, b, c):
		for l in self.__loggers:
			l.__exit__(a, b, c)

class LoggedObject:

	def __init__ (self):
		self.__loggers = []

	@typecheck
	def add_logger (self, logger: Logger):
		self.__loggers.append (logger)

	def log (self, event_type, *k):
		assert issubclass (event_type, LogEvent)

		ev = event_type (self, *k)

		# current thread's logger
		log_event(ev) # FIXME: will raise an error if the thread is not derived from snapshot.Thread

		# object-specific loggers
		for l in self.__loggers:
			l.log_event (ev)

class EventText (metaclass = LogEventClass):
	fields = (("txt", str),)
	def summary (self):
		return self[0]

class EventStep (metaclass = LogEventClass):
	fields = (("name", str),
		  ("description", str))
	def summary (self):
		return "Step %s%s%s" % (
				self[0],
				" - " if self[1] else "",
				self[1]
		)

@typecheck
def log_event (event: LogEvent):
	threading.current_thread().log_event(event) # FIXME: will raise an error if the thread is not derived from snapshot.Thread

def log (*k):
	log_event (EventText (" ".join(str (v) for v in k)))

def step (name, *k):
	log_event (EventStep (str (name), " ".join(str (v) for v in k)))




#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
Analyzer module, defines all the needed tools to analyse a given dump file for
interoperability testing
"""

import glob
import inspect
import sys
import traceback

from os import path
from importlib import import_module

from ttproto.core.data import Data, DifferenceList, Value
from ttproto.core.dissector import (Frame, Capture, is_protocol,
                                    ProtocolNotFound)
from ttproto.core.exceptions import Error
from ttproto.core.typecheck import typecheck, tuple_of, optional, anything, list_of
from ttproto.core.lib.all import *
from ttproto.core.lib.readers.yaml import YamlReader

__all__ = [
    'Verdict',
    'Node',
    'Conversation',
    'TestCase',
    'Analyzer'
]

TESTCASES_SUBDIR = 'testcases'
TTPROTO_DIR = 'ttproto'
TC_FILE_EXTENSION = '.py'
EVERY_TC_WILDCARD = 'td_*' + TC_FILE_EXTENSION


@typecheck
def is_verdict(arg) -> bool:
    """
    Check if a parameter is a valid verdict.
    This function is used for the typechecker decorator.

    :return: True if a valid verdict, False if not
    :rtype: bool
    """
    return all((
        arg is not None,
        type(arg) == str,
        arg in Verdict.values()
    ))


@typecheck
def is_traceback(arg) -> bool:
    """
    Check if a parameter is a valid traceback object.
    This function is used for the typechecker decorator.

    :return: True if a valid traceback, False if not
    :rtype: bool

    .. note:: Didn't find a better way to check this, isinstance or type
              seems to not work
    """
    return all((
        arg is not None,
        hasattr(arg, '__class__'),
        hasattr(arg.__class__, '__name__'),
        isinstance(arg.__class__.__name__, str),
        arg.__class__.__name__ == 'traceback'
    ))


@typecheck
def is_tc_subclass(arg) -> bool:
    """
    Check if a parameter is a valid traceback object.
    This function is used for the typechecker decorator.

    :return: True if a valid traceback, False if not
    :rtype: bool

    .. note:: Didn't find a better way to check this, isinstance or type
              seems to not work
    """
    return all((
        arg is not None,
        type(arg) == type,
        inspect.isclass(arg) and issubclass(arg, TestCase)
    ))


class Verdict:
    """
    A class handling the verdict for an analysis

    Known verdict values are:
     - 'none': No verdict set yet
     - 'pass': The NUT fulfilled the test purpose
     - 'inconclusive': The NUT did not fulfill the test purpose but did not display
                 bad behaviour
     - 'fail': The NUT did not fulfill the test purpose and displayed a bad
               behaviour
     - 'aborted': The test execution was aborted by the user
     - 'error': A runtime error occured during the test

    At initialisation time, the verdict is set to None. Then it can be updated
    one or multiple times, either explicitly calling set_verdict() or
    implicitly if an unhandled exception is caught by the control module
    (error verdict) or if the user interrupts the test manually (aborted
    verdict).

    Each value listed above has precedence over the previous ones. This means
    that when a verdict is updated, the resulting verdict is changed only if
    the new verdict is worse than the previous one.
    """

    __values = ('none', 'pass', 'inconclusive', 'fail', 'aborted', 'error')

    @typecheck
    def __init__(self, initial_value: optional(str) = None):
        """
        Initialize the verdict value to 'none' or to the given value

        :param initial_value: The initial value to put the verdict on
        :type initial_value: optional(str)
        """
        self.__value = 0
        self.__message = ''
        self.__traceback = []
        if initial_value is not None:
            self.update(initial_value)

    @typecheck
    def update(self, new_verdict: str, message: str = ''):
        """
        Update the verdict

        :param new_verdict: The name of the new verdict value
        :param message: The message associated to it
        :type new_verdict: str
        :type message: str
        """
        assert new_verdict in self.__values

        self.__traceback.append((new_verdict, message))
        new_value = self.__values.index(new_verdict)
        if new_value > self.__value:
            self.__value = new_value
            self.__message = message

    @classmethod
    @typecheck
    def values(cls) -> tuple_of(str):
        """
        List the known verdict values

        :return: The known verdict values
        :rtype: (str)
        """
        return cls.__values

    @typecheck
    def get_value(self) -> str:
        """
        Get the value of the verdict

        :return: The value of the verdict as a string
        :rtype: str
        """
        return self.__values[self.__value]

    @typecheck
    def get_message(self) -> str:
        """
        Get the last message update of this verdict

        :return: The last message update
        :rtype: str
        """
        return self.__message

    def get_traceback(self):
        """
        Get the complete traceback of all partial verdicts

        :return: list of partial verdicts, with their messages
        :rtype: list of tuples: [(str, str)]
        """
        return self.__traceback

    @typecheck
    def __str__(self) -> str:
        """
        Get the value of the verdict as string for printing it

        :return: The value of the verdict as a string
        :rtype: str
        """
        return self.__values[self.__value]


class Node:
    """
    A node object is any communicating entity, taking part
    on a conversation. It can be a node, a client/server, an host, etc.

    It has a name (ex: client, node1, iut2, border_router) and an
    associated template, like an IP or a MAC address template for example.
    """

    @typecheck
    def __init__(self, name: str, value: Value):
        """
        Create an node object

        :param name: The name of this node
        :param value: The template associated to this node
        :type name: str
        :type value: Value
        """
        self._name = name
        self._value = value

    def __repr__(self):
        """
        Get the string representation of this node

        :return: String representation of this node
        :rtype: str
        """
        return "Node \"%s\" identification pattern is: %s" % (self._name, self._value)

    @property
    def name(self):
        """
        Function to get the name

        :return: The name
        :rtype: str
        """
        return self._name

    @property
    def value(self):
        """
        Function to get the template

        :return: The template
        :rtype: Value
        """
        return self._value


class Conversation(list):
    """
    A class representing a conversation. A conversation is an ordered exchange
    of messages between at least two nodes.

    It is composed by the nodes and a list of Frames.
    """

    @typecheck
    def __init__(self, nodes: list_of(Node)):
        """
        Function to initialize a Conversation object with its nodes

        :param nodes: The communicating nodes
        :type nodes: [Node]

        :raises ValueError: If not enough nodes received (less than 2)
        """

        # If not enough nodes
        if len(nodes) < 2:
            raise ValueError(
                'At least 2 nodes are required but %d received'
                %
                len(nodes)
            )

        # The node dictionary
        self._nodes = {}
        for node in nodes:
            self._nodes[node.name] = node.value

    @property
    def nodes(self):
        """
        Function to get the nodes as a dictionnary

        :return: The dictionnary of nodes
        :rtype: {str: Template}
        """
        return self._nodes

    def __bool__(self):
        """
        Function for when we check that the conversation exist

        :return: Always True
        :rtype: bool
        """
        return True


class TestCase(object):
    """
    A class handling a test case for an analysis.
    Test cases in ttproto context is a set of checks steps of a test specification
    """

    class Stop(Exception):
        """
        Exception thrown when the execution is finished, can finish sooner than
        the end of the TC if it failed
        """
        pass

    @typecheck
    def __init__(self, capture: Capture):
        """
        Initialize a test case, only input needed is a list of frames

        :param conv_list: The list of conversations to analyze
        :type conv_list: [Conversation]
        """

        # Initialize its verdict instance and its list of conversations
        self._verdict = Verdict()

        # Prepare the parameters
        self._capture = capture
        self._conversations = []
        self._ignore_frames = []
        self._nodes = None
        self._iter = None
        self._frame = None

        # Prepare the values to return after a TC is finished
        # TODO keep & use test_stack (structured representation of the logs in an array) and delete self._text (string)
        self._text = ''
        self._text_stack = []
        self._failed_frames = []
        self._exceptions = []

    @typecheck
    def __not_matching(self, verdict: optional(is_verdict), message: str) -> bool:

        # Put the verdict
        if verdict:
            self.set_verdict(verdict, message)

        # Add this frame's id to the failed frames
        self._failed_frames.append(self._frame['id'])
        # self.log('failed frames: %d' % self._frame['id'])
        self.log(message)
        # Always return False
        return False

    @typecheck
    def match(
            self,
            node_name: optional(str),
            template: Value,
            on_mismatch_verdict: optional(is_verdict) = 'inconclusive',
            on_mismatch_msg: str = ''
    ) -> bool:
        """
        Abstract function to match the current frame value with the template passed as argument

        :param node_name: The node source of the frame
        :param template: The template to match against the frame value
        :param on_mismatch_verdict: Verdict applied when there's a mismatch, normally used with inconclusive, fail
        or None
        :param on_mismatch_msg: Message applied when there's a mismatch
        :type node_name: str
        :type template: Value
        :type on_mismatch_verdict: str
        :type on_mismatch_msg: str

        :return: True if the current frame value matched the given template
                 False if not
        :rtype: bool
        """

        # If no more frames for this conversation
        if not self._iter:
            return self.__not_matching(
                on_mismatch_verdict,
                'Expected %s from the %s but premature end of conversation' % (
                    template, node_name)
            )

        # Check the node
        if node_name is not None:
            try:
                node_template = self._nodes[node_name]

            except KeyError:
                return self.__not_matching(
                    on_mismatch_verdict,
                    'No node %s was not found. Check list of nodes defined for the test case'
                    % (node_name)
                )

            # check the sender (node) is as expected
            try:
                if not node_template.match(self._frame[node_template.__class__]):
                    # The node isn't matching
                    return self.__not_matching(
                        on_mismatch_verdict,
                        'Sender doesnt match. Expected %s pattern for the %s'
                        % (node_template, node_name)
                    )
            except ProtocolNotFound:
                return self.__not_matching(
                    on_mismatch_verdict,
                    'Expected %s into protocol %s but it was not found'
                    % (node_template, node_template.__class__.__name__)
                )

        # Here check the template passed
        protocol = self.get_protocol()
        diff_list = DifferenceList(self._frame[protocol])

        if template.match(self._frame[protocol], diff_list):  # it matches
            if on_mismatch_verdict is not None:
                partial_verdict_message = ' Match: %s' % template
                self.set_verdict('pass', str(self._frame) + partial_verdict_message)

        else:  # mismatch
            if on_mismatch_verdict is not None:
                def callback(path, mismatch, describe):
                    self.log("             %s: %s\n" % (".".join(path), type(mismatch).__name__))
                    # for i in diff_list:
                    #     self.log("             List Diff Item: %s\n" % (i))
                    self.log("                 got:      %s\n" % mismatch.describe_value(describe))
                    self.log("                 expected: %s\n" % mismatch.describe_expected(describe))

                if on_mismatch_msg != '':
                    partial_verdict_message = on_mismatch_msg
                else:
                    partial_verdict_message = ' Mismatch: %s' % template

                # Put the verdict
                self.set_verdict(on_mismatch_verdict, partial_verdict_message)
                diff_list.describe(callback)

            # Add this frame's id to the failed frames
            self._failed_frames.append(self._frame['id'])

            # Frame value didn't match the template
            return False

        # If it matched, return True
        return True

    @typecheck
    def next(self, optional: bool = False):
        """
        Switch to the next frame

        :param optional: If we have to get a next frame or not
        :type optional: bool
        """
        try:
            self._frame = next(self._iter)
            self.log(self._frame)

        except StopIteration:
            if not optional:
                self._iter = None
                self.log('<Frame  ?>')
                self.set_verdict('inconclusive', 'premature end of conversation')

        except TypeError:
            raise self.Stop()

    @typecheck
    def log(self, msg: anything):
        """
        Log a message

        :param msg: The message to log, can be of any type
        :type msg: object
        """
        text = str(msg)
        self._text_stack.append(msg)
        self._text += text if text.endswith('\n') else (text + '\n')

    @typecheck
    def set_verdict(self, verdict: is_verdict, msg):
        """
        Update the current verdict of the current test case

        :param verdict: The new verdict
        :param msg: The message to associate with the verdict

        """
        msg_fix = ''
        if msg is None:
            pass
        elif type(msg) is str:
            msg_fix = msg
            self.log('  [%s] %s' % (format(verdict, "^6s"), msg_fix))
        elif type(msg) is list:
            while msg:
                msg_fix += msg.pop() + '\n'

        # verdict msg can now log more that one line of messages
        self._verdict.update(verdict, msg_fix)

    @typecheck
    def run_test_case(self) -> (
            str,
            list_of(int),
            str,
            list_of((str, str)),
            list_of((type, Exception, is_traceback))
    ):
        """
        Run the test case

        :return: A tuple with the informations about the test results which are
                 - The verdict as a string
                 - The list of the result important frames
                 - A string with the log of the test case
                 - A list of all the partial verdicts and their messages
                 - A list of typles representing the exceptions that occurred
        :rtype: (str, [int], str,[(str,str)], [(type, Exception, traceback)])
        """

        # Pre-process / filter conversations corresponding to the TC
        self._conversations, self._ignored = self.preprocess(self._capture)

        # print("----conversations----")
        # print(self._conversations)
        # print("----ignored----")
        # print(self._ignored)

        # Run the test case for every conversations
        for conv in self._conversations:

            try:

                # Get an iterator on the current conversation frames
                # and its list of nodes
                self._iter = iter(conv)
                self._nodes = conv.nodes
                self.next()

                # Run the test case
                self.run()

            except self.Stop:
                # Ignore this testcase result if the first frame gives an
                # inconclusive verdict
                if all((
                            self._verdict.get_value() == 'inconclusive',
                            self._frame == conv[0]
                )):
                    self.set_verdict('none', 'no match')

            except Exception:

                # Get the execution information, it's a tuple with
                #     - The type of the exception being handled
                #     - The exception instance
                #     - The traceback object
                _exception_type, _exception_value, _exception_traceback = sys.exc_info()

                # Add those exception information to the list
                self._exceptions.append((
                    _exception_type,
                    _exception_value,
                    _exception_traceback
                ))

                # Put the verdict and log the exception
                self.set_verdict('error', 'unhandled exception')
                self.log(_exception_value)

        # Return the results
        return (
            self._verdict.get_value(),
            self._failed_frames,
            self._text,
            self._verdict.get_traceback(),
            self._exceptions,
        )

    @classmethod
    @typecheck
    def get_test_purpose(cls) -> str:
        """
        Get the purpose of this test case

        :return: The purpose of this test case
        :rtype: str
        """
        if cls.__doc__:
            # Get the Yaml reader
            yaml_reader = YamlReader(cls.__doc__, raw_text=True)

            # Then get the dictionnary representation of the tc documentation
            doc_as_dict = yaml_reader.as_dict

            # Into this dict, get the test objective
            assert len(doc_as_dict) == 1
            return doc_as_dict[cls.__name__]['obj']

        return ''

    @classmethod
    @typecheck
    def get_protocol(cls) -> is_protocol:
        """
        Get the protocol corresponding to this test case. This has to be
        implemented into the protocol's common test case class.

        This protocol's layer will be the one on which we will do the matching
        so it should be the lowest one that we are testing.

        :return: The protocol on which this TC will occur
        :rtype: Value
        """
        raise NotImplementedError()

    @typecheck
    def preprocess(
            self,
            capture: Capture
    ) -> (list_of(Conversation), list_of(Frame)):
        """
        Pre-process and filter the frames of the capture into test case related
        conversations. This has to be implemented into the protocol's common
        test case class.
        """
        raise NotImplementedError()

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        raise NotImplementedError()

    @classmethod
    @typecheck
    def get_nodes_identification_templates(cls) -> list_of(Node):
        """
        Get the nodes of this test case. This has to be be implemented into
        each test cases class.

        :return: The nodes of this TC
        :rtype: [Node]
        """
        raise NotImplementedError()


class Analyzer:
    """
        Class for the analyzer tool
    """

    @typecheck
    def __init__(self, test_env: str):
        """
        Initialization function for the analyzer, just fetch the test env
        in which we will get the test description's implementation

        :param test_env: The test environment which is the package name
        :type test_env: str

        :raises NotADirectoryError: If the test environemnt isn't found
        """

        # Check the test_env passed
        test_dir = path.join(
            TTPROTO_DIR,
            test_env,
            TESTCASES_SUBDIR
        )
        if not path.isdir(test_dir):
            raise NotADirectoryError(
                "The test cases dir %s wasn't found"
                %
                test_dir
            )

        # LoggedObject.__init__(self)
        self.__test_env = test_env

    @typecheck
    def __fetch_from_pathname(self, testcases: list_of(type), search: str):
        """
        Fetch test cases from the test suite plugin

        :param testcases: List in which we will add the test cases
        :param search: The research query, can be a single TC or the wildcard
                       to select all the test cases
        :type testcases: [type]
        :type search: str
        """

        # Build the search query
        search_query = path.join(
            TTPROTO_DIR,
            self.__test_env,
            TESTCASES_SUBDIR,
            search
        )

        # Fetch files using the search query
        result = glob.glob(search_query)

        # If no test case found
        if len(result) == 0:
            raise FileNotFoundError(
                'No test case found for: "%s"' % search_query
            )

        # If the search query is the wildcard, sort the list
        elif search == EVERY_TC_WILDCARD:
            result.sort()

        # For every file found
        for filepath in result:
            # Get the name of the file
            filename = path.basename(filepath)

            # Get the name of the module
            modname = filename[:(-1 * len(TC_FILE_EXTENSION))]

            # Build the module relative name
            mod_rel_name = '.'.join([
                TTPROTO_DIR,
                self.__test_env,
                TESTCASES_SUBDIR,
                modname
            ])

            # Note that the module is always lower case and the plugin (class)
            # is upper case (ETSI naming convention)
            tc_class = getattr(
                import_module(mod_rel_name),
                modname.upper()
            )
            testcases.append(tc_class)

    @typecheck
    def import_test_cases(
            self,
            testcases: optional(list_of(str)) = None
    ) -> list:
        """
        Imports test cases classes from TESTCASES_DIR

        :param testcases: The wanted test cases as a list of string
        :type testcase_id: optional([str])

        :return: List of test cases class in the same order than the param list
        :rtype: [TestCase]

        :raises FileNotFoundError: If no test case was found

        .. note::
            Assumptions are the following:
                - Test cases are defined inside a file, each file contains
                  only one test case
                - All test cases must be named td_*
                - All test cases are contained into ttproto/[env]/testcases
                - Filenames corresponds to the TC id in lower case
                - Class names corresponds to the TC id
        """

        # The return values
        tc_fetched = []

        # If no TCs provided, fetch all the test cases found
        if not testcases:
            self.__fetch_from_pathname(tc_fetched, EVERY_TC_WILDCARD)

        # If testcases list are provided, fetch those
        else:

            # For every test case given
            for test_case_name in testcases:
                tc_name_query = test_case_name.lower() + TC_FILE_EXTENSION
                self.__fetch_from_pathname(tc_fetched, tc_name_query)

        # Return the test cases classes
        return tc_fetched

    @typecheck
    def get_implemented_testcases(
            self,
            testcases: optional(list_of(str)) = None,
            verbose: bool = False
    ) -> list_of((str, str, str, str)):
        """
        Get more informations about the test cases

        :param testcases: A list of test cases to get their informations
        :param verbose: True if we want more informations about the TC
        :type testcase_id: optional([str])
        :type verbose: bool

        :raises FileNotFoundError: If one of the test case is not found

        :return: List of descriptions of test cases composed of:
                    - tc_identifier
                    - tc_objective
                    - tc_sourcecode
                    - tc_doc
        :rtype: [(str, str, str, str)]
        """

        # The return value
        ret = []

        # Get the tc classes
        tc_classes = self.import_test_cases(testcases)

        # Add the infos of each of them to the return value
        for tc in tc_classes:

            # If verbose is asked, we provide the source code and doc too
            source_code = ''
            source_doc = ''
            if verbose:
                source_code = inspect.getsource(tc)
                source_doc = inspect.getdoc(tc)

            # Add the tuple to the return value
            ret.append(
                (tc.__name__, tc.get_test_purpose(), source_code, source_doc)
            )

        # Return the list of tuples
        return ret

    @typecheck
    def analyse(
            self,
            filename: str,
            tc_id: str
    ) -> (str, str, list_of(int), str, list_of((str, str)), list_of((type, Exception, is_traceback))):
        """
        Analyse a dump file associated to a test case

        :param filename: The name of the file to analyse
        :param tc_id: The unique id of the test case to confront the given file
        :type filename: str
        :type tc_id: str

        :return: A tuple with the information about the analysis results:
                 - The id of the test case
                 - The verdict as a string
                 - The list of the result important frames
                 - A string with logs
                 - A list of all the partial verdicts
                 - A list of tuples representing the exceptions that occurred
        :rtype: (str, str, [int], str,[(str, str)], [(type, Exception, traceback)])

        :raises FileNotFoundError: If the test env of the tc is not found
        :raises ReaderError: If the capture didn't manage to read and decode
        :raises ObsoleteTestCase: If the test case if obsolete

        .. example::
            ('TD_COAP_CORE_03', 'fail', [21, 22], [('fail', "some comment"),('fail', "some other comment")] , 'verdict description', '')

        .. note::
            Allows multiple occurrences of executions of the testcase, returns as verdict:
                - fail: if at least one on the occurrences failed
                - inconclusive: if all occurrences returned a inconclusive verdict
                - pass: all occurrences are inconclusive or at least one is PASS and
                        the rest is inconclusive
        """

        # Get the test case class
        test_case_class = self.import_test_cases([tc_id])
        assert len(test_case_class) == 1
        test_case_class = test_case_class[0]

        # Disable name resolution for performance improvment
        with Data.disable_name_resolution():
            # Get the capture from the file
            capture = Capture(filename)

            # Initialize the TC with the list of conversations
            test_case = test_case_class(capture)
            verdict, rev_frames, log, partial_verdicts, exceptions = test_case.run_test_case()

            # print('##### Ignored')
            # print(ignored)
            # print('#####')

            # Here we execute the test case and return the result

            # print('##### Verdict given')
            # print(verdict)
            # print('#####')
            # print('##### Review frames')
            # print(rev_frames)
            # print('#####')
            # print('##### Text')
            # print(extra)
            # print('#####')
            # print('##### Exceptions')
            # print(exceptions)
            # print('#####')

            # Return the result

            return tc_id, verdict, rev_frames, log, partial_verdicts, exceptions


if __name__ == "__main__":
    from os import getcwd, path

    #
    # analyzer = Analyzer('tat_6lowpan')
    #
    # params = './tmp/TD_6LOWPAN_HC_01.pcap', 'TD_6LOWPAN_HC_01'
    # tc_name, verdict, rev_frames, str_log, lst_log, excepts = analyzer.analyse(params[0], params[1])
    # print('##### TC name')
    # print(tc_name)
    # print('#####')
    # print('##### Verdict given')
    # print(verdict)
    # print('#####')
    # print('##### Review frames')
    # print(rev_frames)
    # print('#####')
    # print('##### Text')
    # print(str_log)
    # print('##### Partial verdicts')
    # for s in lst_log:
    #     print(str(s))
    # print('#####')
    # print('##### Exceptions')
    # for e in excepts:
    #     e1, e2, e3 = e
    #     print(repr(traceback.format_exception(e1, e2, e3)))
    #
    # print('#####')


    analyzer = Analyzer('tat_coap')

    # params = './tests/test_dumps/AnalyzerTests/coap_core/TD_COAP_CORE_01_pass.pcap', 'TD_COAP_CORE_01'
    # params = './tests/test_dumps/analysis/coap_core/TD_COAP_CORE_01_pass.pcap', 'TD_COAP_CORE_01'
    params = './tests/test_dumps/analysis/coap_core/TD_COAP_CORE_03_FAIL_No_CoAPOptionContentFormat.pcap', 'TD_COAP_CORE_03'
    # params = './tmp/TD_COAP_CORE_23_fail.pcap', 'TD_COAP_CORE_23'
    # params = './tests/test_dumps/AnalyzerTests/coap_core/TD_COAP_CORE_03_FAIL_No_CoAPOptionContentFormat.pcap', 'TD_COAP_CORE_03'
    tc_name, verdict, rev_frames, str_log, lst_log, excepts = analyzer.analyse(params[0], params[1])
    print('##### TC name')
    print(tc_name)
    print('#####')
    print('##### Verdict given')
    print(verdict)
    print('#####')
    print('##### Review frames')
    print(rev_frames)
    print('#####')
    print('##### Text')
    print(str_log)
    print('##### Partial verdicts')
    for s in lst_log:
        print(str(s))
    print('#####')
    print('##### Exceptions')
    for e in excepts:
        e1, e2, e3 = e
        print(repr(traceback.format_exception(e1, e2, e3)))

    print('#####')



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os, time , sys

sys.path.append('..')

from	ttproto.core.lib.all		import *
from	ttproto.core.port		import *

from	ttproto.core.logger		import *
from	ttproto.core.html_logger		import *
from	ttproto.core.control		import *
from	ttproto.core.data		import *
from	ttproto.core.snapshot		import *
from	ttproto.core.lib.ports.socat	import *
from	ttproto.core.exceptions		import TerminateTestcase

from	.config			import *
from 	.altsteps		import *
from 	.commontestsetups	import *

from ttproto.core.list			import *
from ttproto.core.templates 		import *

#FIXME: move those functions in a specific file

import fcntl, select
# make stdin a non-blocking file
stdin_fd = sys.stdin.fileno()
fcntl.fcntl(stdin_fd, fcntl.F_SETFL, fcntl.fcntl(stdin_fd, fcntl.F_GETFL) | os.O_NONBLOCK)


# FIXME : we will need a way to a add callback function, or to grab the data, at least for Test_LPND_1_1_6d,f
def receive_unordered (link, *datas):
	data_list = list (datas)
	received_values = []
	while data_list:
		with alt:
			for d in data_list:
				@link.receive (d)
				def _(data, value):
					try:
						data_list.remove(data)
					except ValueError:
						data_list.remove(data["pl"]["pl"]) # FIXME: will raise if receiving explicitely Sixlowpan(IPv6|IPHC) and/or if the data is not flat
						
					received_values.append(value)
					set_verdict ("pass")
	return received_values

def superop ( opt ):
	return Superset (ICMPv6OptionList , opt)



import unittest, logging, json

from os import getcwd, path
from ttproto.core.analyzer import Analyzer
from ttproto.core.typecheck3000 import InputParameterError
from tests.test_tools.struct_validator import StructureValidator


class CoAPAnalyzerTestCase(unittest.TestCase):
    """
    Test class for the analyzer tool
    """

    # #################### Tests parameters #########################

    # Test env (only tat_coap for the moment)
    TEST_ENV = 'tat_coap'
    TEST_DIR = './tests/test_dumps/analysis/coap_core/'
    UNKNOWN_TEST_ENV = 'unknown'
    TEST_CASE_ID = 'TD_COAP_CORE_01'
    TEST_CASE_ID_WHICH_BUGGED_IN_THE_PAST = 'TD_COAP_CORE_24'
    UNKNOWN_TEST_CASE_ID = 'TD_COAP_CORE_42'

    # Create a struct checker object
    struct_validator = StructureValidator()

    # #################### Init and deinit functions #########################
    def setUp(self):
        """
            Initialize the analyzer instance
        """
        self.analyzer = Analyzer(self.TEST_ENV)

    # #################### Tests functions #########################

    # ##### __init__
    def test___init__(self):

        # Initialize the analyzer with a correct test env
        analyzer = Analyzer(self.TEST_ENV)

    def test___init___unknown_test_env(self):

        # Initialize the analyzer with an unknown test env
        with self.assertRaises(NotADirectoryError):
            analyzer = Analyzer(self.UNKNOWN_TEST_ENV)

    # ##### get_implemented_testcases
    def test_get_implemented_testcases(self):

        # Get implemented test cases and check their values
        tcs = self.analyzer.get_implemented_testcases()
        self.struct_validator.check_tc_from_analyzer(tcs)

    def test_get_implemented_testcases_with_none_value(self):

        # Get implemented test cases and check their values
        tcs = self.analyzer.get_implemented_testcases(None)
        self.struct_validator.check_tc_from_analyzer(tcs)

    def test_get_implemented_testcases_single_test_case(self):

        # Get implemented test cases and check their values
        tcs = self.analyzer.get_implemented_testcases([self.TEST_CASE_ID])
        self.struct_validator.check_tc_from_analyzer(tcs)
        self.assertEqual(len(tcs), 1)
        self.assertEqual(tcs[0][0], self.TEST_CASE_ID)

    def test_get_implemented_testcases_verbose_mode(self):

        # Get implemented test cases and check their values
        tcs = self.analyzer.get_implemented_testcases(verbose=True)
        self.struct_validator.check_tc_from_analyzer(tcs)

        # Check that they have the extra informations (the source code)
        for tc in tcs:
            self.assertGreater(len(tc[2]), 0)

    def test_get_implemented_testcases_single_test_case_which_bugged(self):

        if self.TEST_CASE_ID_WHICH_BUGGED_IN_THE_PAST:
            # Get implemented test cases and check their values
            tcs = self.analyzer.get_implemented_testcases(
                [self.TEST_CASE_ID_WHICH_BUGGED_IN_THE_PAST]
            )
            self.struct_validator.check_tc_from_analyzer(tcs)
            self.assertEqual(len(tcs), 1)
            self.assertEqual(tcs[0][0], self.TEST_CASE_ID_WHICH_BUGGED_IN_THE_PAST)

    def test_get_implemented_testcases_unknown_test_case(self):

        # Get implemented test cases and check their values
        with self.assertRaises(FileNotFoundError):
            tcs = self.analyzer.get_implemented_testcases(
                [self.UNKNOWN_TEST_CASE_ID]
            )

    def test_get_implemented_testcases_str_instead_of_list(self):

        # Get implemented test cases and check their values
        with self.assertRaises(InputParameterError):
            tcs = self.analyzer.get_implemented_testcases(self.TEST_CASE_ID)

    # ##### analyse
    def test_analyse_basic_pass_PCAPs(self):
        dir = self.TEST_DIR
        print('looking for test dumps for testing the test cases: %s' % dir)
        for tc in self.analyzer.get_implemented_testcases():
            filename = path.join(dir, tc[0] + '_PASS.pcap')
            print('Testcase found %s , dump file %s for test exist: %s' % (str(tc[0]), filename, path.isfile(filename)))
            # check if there's a pcap_pass_test for the testcase
            if path.isfile(filename):
                tc_name, verdict, rev_frames, log, partial_verdicts, exception_info = self.analyzer.analyse(filename,
                                                                                                            tc[0])

                # I apologize to you future reader for the utterly ugly way of unpacking the exception info :)
                exc_str = ''
                if exception_info:
                    counter = 0
                    for i in exception_info:
                        counter += 1
                        exc_str += '\nException n:%s' % counter
                        exc_str += '\n\ttype: %s ' % i[0]
                        exc_str += '\n\texception: %s' % i[1]
                        exc_str += '\n\ttracelog: %s' % i[2]
                        exc_str += '\n'

                self.assertTrue(verdict == 'pass',
                                msg='%s implementation got verdict: %s expected PASS \n '
                                    'details: \n %s \n'
                                    'errors: \n %s \n'
                                    'logs: \n %s \n'
                                    %
                                    (
                                        tc_name,
                                        str(verdict).upper(),
                                        json.dumps(partial_verdicts, indent=4) if log else "",
                                        exc_str if exc_str else "No exception registered",
                                        log
                                    )
                                )

                print('Testcase %s , got verdict: %s' % (str(tc[0]), str(verdict).upper()))


class SixlowpanHcAnalyzerTestCase(CoAPAnalyzerTestCase):
    # #################### Tests parameters #########################
    TEST_ENV = 'tat_6lowpan'
    TEST_DIR = './tests/test_dumps/analysis/6lowpan_hc/'
    UNKNOWN_TEST_ENV = 'unknown'
    TEST_CASE_ID = 'TD_6LOWPAN_HC_01'
    UNKNOWN_TEST_CASE_ID = 'TD_6LOWPAN_HC_666'
    TEST_CASE_ID_WHICH_BUGGED_IN_THE_PAST = None


# #################### Main run the tests #########################
if __name__ == '__main__':
    unittest.main()



# Invokes webserver to be run at 127.0.0.1:2080 if INTERFACE is amqp,
# else runs amqp interface
# Should be run as: python3 -m ttproto.tat_coap

import sys
import os
import logging
import argparse
from ttproto import LOG_LEVEL
from ttproto.tat_services import dissect_capture, analyze_capture, get_protocols_list, ALLOWED_PROTOCOLS_FOR_ANALYSIS
from multiprocessing import Process

# TTPROTO CONSTANTS
COMPONENT_ID = 'tat|main'
SERVER_CONFIG = ("0.0.0.0", 2080)

# default handler
logger = logging.getLogger(COMPONENT_ID)
logger.setLevel(LOG_LEVEL)

PCAP_DUMPER_AMQP_TOPICS = [
    'fromAgent.#.packet.raw'
    'data.*.fromAgent.*',  # fixme deprecate this (old API)
]

cli_help = '''ttproto <command> [<args>]

TTProto CLI accepts the following commands:
    dissect         Dissects network traces (.pcap file).
    analyze         Analyses network traces (.pcap file).
    service_amqp    Launches TTProto as a HTTP service (WIP).
    service_http    Launches TTProto as an AMQP service.
'''
diss_help = ''' ttproto dissect file [<options>]

Dissection usage examples:
    dissect ./tests/test_dumps/dissection/6lowpan/TD_6LOWPAN_HC_01.pcap
    dissect ./tests/test_dumps/dissection/6lowpan/TD_6LOWPAN_HC_01.pcap -o /tmp/dissection.json
    dissect ./tests/test_dumps/dissection/6lowpan/TD_6LOWPAN_HC_01.pcap -p sixlowpan
    dissect ./tests/test_dumps/dissection/6lowpan/TD_6LOWPAN_HC_01.pcap -p icmpv6
    dissect ./tests/test_dumps/dissection/6lowpan/TD_6LOWPAN_HC_01.pcap -p icmpv6echorequest

'''
ana_help = '''ttproto analyze file [<options>]

Analyze usage examples:
    analyze ./tests/test_dumps/dissection/6lowpan/TD_6LOWPAN_HC_01.pcap -p 6lowpan -tc TD_6LOWPAN_HC_01 
    analyze ./tests/test_dumps/analysis/coap_core/TD_COAP_CORE_01_PASS.pcap -p coap -tc TD_COAP_CORE_01
'''


class TTProtoCLI():
    def __init__(self):
        parser = argparse.ArgumentParser(
            description='Interface for TTProto.',
            usage=cli_help
        )
        parser.add_argument('command', help='Subcommand to run')
        # parse_args defaults to [1:] for args, but you need to
        # exclude the rest of the args too, or validation will fail
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            logging.error('Unrecognized command')
            parser.print_help()
            exit(1)
        # use dispatch pattern to invoke method with same name
        getattr(self, args.command)()

    def dissect(self):
        # parse arguments
        parser = argparse.ArgumentParser(
            description=dissect_capture.__doc__,
            usage=diss_help
        )
        parser.add_argument('file',
                            help="Filename (complete path) to traces file (.pcap file).",
                            )
        parser.add_argument("-p", "--protocol",
                            choices=[i.lower() for i in get_protocols_list()],
                            help="Choose a filter protocol",
                            )
        parser.add_argument("-o", "--output",
                            help="Output file name",
                            default='dissection.json',
                            )

        try:
            args = parser.parse_args(sys.argv[2:])
        except Exception as e:
            print(e)
            parser.print_help()
            exit(1)

        _, disects_as_list_of_text = dissect_capture(
            filename=args.file,
            proto_filter=args.protocol,
            output_file=args.output
        )

        if disects_as_list_of_text:
            for item in disects_as_list_of_text:
                logger.info(item)
                # print(item.encode('utf-8').decode('unicode_escape')) this doesnt work very well for some bytes

    def analyze(self):
        # parse arguments
        parser = argparse.ArgumentParser(
            description=analyze_capture.__doc__,
            usage=ana_help,
        )

        parser.add_argument('file',
                            help="Filename (complete path) to traces file (.pcap file).",
                            )
        parser.add_argument("-p", "--protocol",
                            choices=ALLOWED_PROTOCOLS_FOR_ANALYSIS,
                            help="Choose the protocol to be analyzed by the TAT.",
                            default='coap',
                            )
        parser.add_argument("-tc", "--test-case",
                            help="Indicate the particular test case ID to check."
                            )
        parser.add_argument("-o", "--output",
                            help="Output file name",
                            default='analysis.json',
                            )
        try:
            args = parser.parse_args(sys.argv[2:])
        except Exception as e:
            print(e)
            parser.print_help()
            exit(1)

        analysis_results = analyze_capture(
            filename=args.file,
            protocol=args.protocol,
            testcase_id=args.test_case,
            output_file=args.output
        )

        if analysis_results:
            logger.info('result: %s' % analysis_results[1])
            logger.info('details: \n%s' % analysis_results[3])

    def service_http(self):
        raise NotImplementedError

        # from ttproto.tat_coap.webserver import http, RequestHandler
        # __shutdown = False
        #
        # def shutdown():
        #     global __shutdown
        #     __shutdown = True
        #
        # server = http.server.HTTPServer(SERVER_CONFIG, RequestHandler)
        # logger.info('Server is ready: %s:%s' % SERVER_CONFIG)
        #
        # while not __shutdown:
        #     try:
        #         server.handle_request()
        #     except Exception as e:
        #         logger.error(str(e))

    def service_amqp(self):

        from ttproto.tat_amqp_interface import launch_tat_amqp_interface, TMPDIR
        from ttproto.utils.packet_dumper import launch_amqp_data_to_pcap_dumper
        from ttproto.utils.rmq_handler import JsonFormatter, RabbitMQHandler

        # parse arguments
        parser = argparse.ArgumentParser(description='Launches TTProto as an AMQP service.')

        parser.add_argument("-p", "--protocol",
                            choices=ALLOWED_PROTOCOLS_FOR_ANALYSIS,
                            help="Choose the protocol to be analyzed by the TAT.",
                            default='coap',
                            )

        parser.add_argument("-d", "--dissector",
                            action='store_true',
                            default=False,
                            help="Launches the dissector component which listens to AMQP bus, "
                                 "dissects all exchanged frames and pushes results back into the bus")

        parser.add_argument("-s", "--dumps",
                            action='store_true',
                            default=False,
                            help="Launches a component which listens to data plane in AMQP bus and dumps traces to "
                                 "pcap file. --dissector flag auto enables this mode.")

        args = parser.parse_args(sys.argv[2:])

        tat_interface = 'AMQP'
        tat_protocol = args.protocol
        dissector_option = args.dissector
        dumps_option = args.dumps

        if dissector_option:  # auto dissection needs the traces in pcap files
            dumps_option = True

        logger.info('Configuration: \n\tinterface: %s\n\tprotocol: %s \n\tauto-dissection option: %s'
                    % (tat_interface, tat_protocol, dissector_option))

        # AMQP ENV variables (either get them all from ENV or set them all as default)
        try:
            amqp_exchange = str(os.environ['AMQP_EXCHANGE'])
        except KeyError:
            logger.warning('Cannot retrieve environment variables for AMQP connection. Loading defaults..')
            amqp_exchange = "amq.topic"

        try:
            amqp_url = str(os.environ['AMQP_URL'])
        except KeyError:
            logger.warning('Cannot retrieve environment variables for AMQP connection. Loading defaults..')
            amqp_url = "amqp://local:{1}@{2}/{3}".format('guest', 'guest', 'localhost', '')

        logger.info('Env vars for AMQP connection succesfully imported: AMQP_URL: %s, AMQP_EXCHANGE: %s' % (amqp_url,
                                                                                                            amqp_exchange))

        # AMQP log handler & formatter
        rabbitmq_handler = RabbitMQHandler(amqp_url, COMPONENT_ID)
        json_formatter = JsonFormatter()
        rabbitmq_handler.setFormatter(json_formatter)
        logger.addHandler(rabbitmq_handler)
        logger.setLevel(logging.INFO)

        logger.info('Starting AMQP interface of TAT')

        # launch process: TAT
        p_tat = Process(target=launch_tat_amqp_interface,
                        args=(amqp_url, amqp_exchange, tat_protocol, dissector_option))
        p_tat.start()

        # launch process: pcap dumper
        if dumps_option or dissector_option:  # dissector component needs dumper
            logger.info('Starting AMQP data plane to pcap dumper')
            p_dumper = Process(
                target=launch_amqp_data_to_pcap_dumper(),
                args=(amqp_url, amqp_exchange, PCAP_DUMPER_AMQP_TOPICS, TMPDIR)
            )
            p_dumper.start()


if __name__ == '__main__':
    logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)
    TTProtoCLI()



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# TODO: maybe some import are unused

from	ttproto.core.data		import Value
from	ttproto.core.union		import *
from	ttproto.core.packet		import *
from	ttproto.core.exceptions		import Error
from	ttproto.core.typecheck		import *
from	ttproto.core.lib.inet.meta	import *
from	ttproto.core.lib.inet.basics 	import *
from	ttproto.core.lib.inet.ipv6	import *
from	ttproto.core.lib.inet.udp	import *
from	ttproto.core.lib.inet.sixlowpan	import *
from	ttproto.core.lib.inet.sixlowpan_hc import *

from	ttproto.core			import port

__all__ = [
	'SixLowpanMessagePort'
	]

class SixLowpanMessagePort (port.MessagePort):

	@typecheck
	def __init__ (self, pan_id: int, *k, compression = False):
		"""
		pan_id		PAN ID used by the port (at the 802.15.4 layer)

		compression	if True, outgoing IPv6 messages will be compressed by default
				(when .send() is called on an IPv6 message w/o any 6lowpan header)
		"""

		super().__init__(*k)

		self.__seq	= 0
		self.__pan_id	= pan_id

		self.compression = compression

		import ttproto.core.lib.ieee802154
		self.Ieee802154 = ttproto.core.lib.ieee802154.Ieee802154

	def __resolve_addr (self, ip6):
		# TODO: support manually-configured addresses

		if ip6[0] == 0xff:
			# multicast address
			if ip6 == IPv6Address ("ff02::1"):
				return b'\xff\xff'
			else:
				# mapping of multicast address (RFC4944 Section 9)
				return bytes((0x80 | (ip6[14] & 0x1f), ip6[15]))

		elif ip6[10:14] == b"\0\xff\xfe\0":
			# short address
			return bytes ((ip6[14], ip6[15]))
		else:
			# EUI-64
			return bytes((ip6[8] ^ 2,)) + ip6[9:]

	def __next_seq (self):
		with self.lock:
			result = self.__seq
			self.__seq += 1
			return result

	def send (self, value , src=None , dst=None ):
		value  = value.flatten()
		ip_hdr = None
		if isinstance (value, IPv6):
			ip_hdr = value
			value = (SixLowpanIPHC() if self.compression else SixLowpanIPv6()) / value
		elif isinstance (value, SixLowpanIPv6):
			ip_hdr = value["pl"]
		elif isinstance (value, SixLowpanIPHC):
			ip_hdr = value["pl"]
		elif isinstance (value, SixLowpanFRAG1):
			if isinstance (value["pl"] , SixLowpanIPv6):
				ip_hdr = value["pl"]["pl"]
		#elif isinstance (value, SixLowpan):
		#	if isinstance (value["pl"] , IPv6):
		#		ip_hdr = value["pl"]

		assert None not in (src, dst) or isinstance (ip_hdr, IPv6) # TODO: support other cases

		if isinstance (ip_hdr, IPv6):
			if not src:
				src = ip_hdr["src"]
			if not dst:
				dst = ip_hdr["dst"]
			if not src:
				src = IPv6Address("::")
			if not dst:
				dst = IPv6Address("::")

		super().send (
			self.Ieee802154 (
				dpid	= self.__pan_id,
				dst	= self.__resolve_addr (dst),
				src	= self.__resolve_addr (src),
				seq	= self.__next_seq(),
				pl	= value,
				ar	= 1,
			)
		)

	def match_receive (self, data: optional (is_data) = None, src=None, dst=None):
		#TODO: match the src/dst addresses
		if data is None:
			assert src is None and dst is None
			return super().match_receive()

		if issubclass (data.get_type(), IPv6):
			#FIXME: this creates an extra mismatch logevent each time the function is called
			for sixlowpan_type in (SixLowpanIPv6, SixLowpanIPHC):
				result = super().match_receive(self.Ieee802154(src=src, dst=dst)/sixlowpan_type(pl=data))
				if result:
					return result
			return None
		else:
			return super().match_receive(self.Ieee802154()/data)



#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *


class TD_COAP_CORE_21(CoAPTestcase):
    """Identifier:
TD_COAP_CORE_21
Objective:
Perform GET transaction containing the ETag option (CON mode)
Configuration:
CoAP_CFG_BASIC
References:
[COAP] 5.8.1, 5.10.7,5.10.10,12.1.12

Pre-test
conditions:
•	Server should offer a /validate resource which vary in time
•	Client & server supports ETag option
•	The Client‘s cache must be purged

Test Sequence:
Step
Type
Description

Part A: Verifying that client cache is empty

1
Stimulus
Client is requested to send a confirmable GET request to server’s resource


2
Check
The request sent request by the client contains:
•	Type = 0 (CON)
•	Code = 1 (GET)
•	Client-generated Message ID (➔ CMID)
•	Client-generated Token (➔ CTOK)
•	Uri-Path option "validate"
•	No ETag option

3
Check
Server sends response containing:
•	Code = 69 (2.05 content)
•	Message ID = CMID, Token = CTOK
•	Option type = ETag, value = a value chosen by the server (➔ ETAG1)
•	Non-empty Payload

4
Verify
Client displays the response

Part B: Verifying client cache entry is still valid

5
Stimulus
Client is requested to send s confirmable GET request to
server’s resource so as to check if the resource was updated

6
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 1 (GET)
•	Another client-generated Message ID ≠ CMID (➔ CMID2)
•	Client-generated Token which may or may not be ≠ CTOK (➔ CTOK2)
•	Uri-Path option "validate"
•	Option Type = ETag, value = ETAG1 (the ETag value received in step 3)

7
Check
Server sends response containing:
•	Code = 67 (2.03 Valid)
•	Message ID = CMID2, Token = CTOK2
•	Option type = ETag, value = ETAG1
•	Empty Payload


8
Verify
Client displays the response

Part C: Verifying that client cache entry is no longer valid

9
Stimulus
Update the content of the server’s resource from a CoAP client (either another client, or the testing client in a separate transaction)

10
Stimulus
Client is requested to send a confirmable GET request to server’s resource so as to check if the resource was updated

11
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 1 (GET)
•	Another client-generated Message ID ≠ CMID and ≠ CMID2 (➔ CMID3)
•	Client-generated Token which may or may not be ≠ CTOK or CTOK2 (➔ CTOK3)
•	Uri-Path option "validate"
•	Option Type = ETag, value = ETAG1 (the ETag value received in step 3)

12
Check
Server sends response containing:
•	Code = 69 (2.05 Content)
•	Message ID = CMID3, Token = CTOK3
•	Option type = ETag, value = another ETag value ≠ ETAG1
•	The payload of the requested resource, which should be different from the payload in step 3


13
Verify
Client displays the response
"""

    def run(self):
        # Part A
        self.match_coap("client", CoAP(type="con", code="get",
                                       opt=All(
                                           Opt(CoAPOptionUriPath("validate")),
                                           NoOpt(CoAPOptionETag()),
                                       )))
        CMID = self.frame.coap["mid"]
        CTOK = self.frame.coap["tok"]

        self.next_skip_ack()

        if not self.match_coap("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                              code=2.05,
                                              mid=CMID,
                                              tok=CTOK,
                                              opt=Opt(CoAPOptionETag()),
                                              pl=Not(b""))):
            raise self.Stop()

        ETAG1 = self.frame.coap["opt"][CoAPOptionETag]["val"]
        pl3 = self.frame.coap["pl"]

        self.next_skip_ack(optional=True)

        # Part B
        self.chain()

        self.match_coap("client", CoAP(type="con", code="get",
                                       opt=Opt(
                                           CoAPOptionUriPath("validate"),
                                           CoAPOptionETag(ETAG1),
                                       )))
        CMID2 = self.frame.coap["mid"]
        CTOK2 = self.frame.coap["tok"]
        if CMID2 is Not(b''):
            if CMID2 == CMID:
                self.setverdict("fail", "Message ID should be different")
        if CTOK2 is Not(b''):
            if CTOK2 == CTOK:
                self.setverdict("fail", "Token should be different")

        self.next_skip_ack()

        self.match_coap("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                       code=2.03,
                                       mid=CMID2,
                                       tok=CTOK2,
                                       opt=Opt(CoAPOptionETag(ETAG1)),
                                       pl=b""))

        self.next_skip_ack(optional=True)

        # Part C
        self.chain()

        if self.match_coap("client", CoAP(code="put"), None):
            # allow an update from another client running on the same host
            self.next_skip_ack()
            self.match_coap("server", CoAP(code=2.04))
            self.next_skip_ack(optional=True)

            self.chain()

        self.match_coap("client", CoAP(type="con", code="get",
                                       opt=Opt(
                                           CoAPOptionUriPath("validate"),
                                           CoAPOptionETag(ETAG1),
                                       )))
        CMID3 = self.frame.coap["mid"]
        CTOK3 = self.frame.coap["tok"]
        if CMID3 is Not(b''):
            if CMID3 == CMID or CMID3 == CMID2:
                self.setverdict("fail", "Message ID should be different")
        if CTOK3 is Not(b''):
            if CTOK3 == CTOK or CTOK3 == CTOK2:
                self.setverdict("fail", "Token should be different")

        self.next_skip_ack()

        self.match_coap("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                       code=2.05,
                                       mid=CMID3,
                                       tok=CTOK3,
                                       opt=Opt(CoAPOptionETag(Not(ETAG1))),
                                       pl=All(Not(b""), Not(pl3))))

        self.next_skip_ack(optional=True)



#!/usr/bin/env python3
#
#   (c) 2012    Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.    You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors    have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import sys, time, re, itertools, socket, urllib.parse, glob, inspect
from lib2to3.fixes.fix_print import parend_expr
from os import chdir, path, getcwd
from importlib import import_module

from ttproto.core.lib.readers.pcap import PcapReader
from ttproto.core.data import Data, Value
from ttproto.core.list import ListValue
from ttproto.core.packet import PacketValue


from ttproto.core.lib.ethernet import *
from ttproto.core.lib.encap import *

import ttproto.core.lib.inet.all
from ttproto.core.lib.inet.ipv4 import *
from ttproto.core.lib.inet.ipv6 import *
from ttproto.core.lib.inet.udp import *
from ttproto.core.lib.inet.coap import *
from ttproto.core.lib.inet.sixlowpan import *
from ttproto.core.lib.inet.sixlowpan_hc import *

from ttproto.core.lib.ieee802154 import *

from ttproto.core.xmlgen import XHTML10Generator
from ttproto.core.html_logger import HTMLLogger
from ttproto.utils.version_git import get_git_version

from collections import OrderedDict


# import the specifics for the protocol to be tested
from . import proto_specific

# small hack to allow participants running a server on non standard ports:
#import ttproto.core.lib.inet.udp
#for port in (5684, 5685, 5686, 5687, 5688, 5689):
#    ttproto.core.lib.inet.udp.udp_port_map[port] = CoAP


TOOL_VERSION = get_git_version()
TEST_VERSION = "td-coap4_&_IRISA"
TESTCASES_SUBDIR = '/ttproto/ts_coap/testcases'

# TODO abstract classes?
TestCase = proto_specific.CoAPTestcase
Tracker = proto_specific.CoAPTracker

class Resolver:
    __cache = {}

    def __new__ (cls, ip_addr):
        try:
            return cls.__cache[ip_addr]
        except KeyError:
            pass

        try:
            name = socket.gethostbyaddr (str (ip_addr))[0]
        except socket.herror:
            name = None

        cls.__cache[ip_addr] = name

        return name

    @classmethod
    def format (cls, ip_addr):
        name = cls (ip_addr)

        if name:
            return "%s (%s)" % (name, ip_addr)
        else:
            return ip_addr

class Frame:
    """

    """
    @classmethod
    def create_list (cls, pcap_frames):
        return list (cls (i, f) for i,f in zip (itertools.count(1), pcap_frames))

    def __init__ (self, id, pcap_frame):
        self.id = id
        self.ts, self.msg, self.exc = pcap_frame
        self.__extract_infos()

    def __repr__ (self):
        return "<Frame %3d: %s>" % (self.id, self.msg.summary())

    def __extract_infos (self):
        self.src    = None
        self.dst    = None
        self.coap   = None

        v = self.msg.get_value()
        while True:
            if isinstance (v, Ethernet) or isinstance (v, IPv6) or isinstance (v, IPv4):
                self.src = v["src"]
                self.dst = v["dst"]
                v = v["pl"]
                continue
            elif isinstance (v, UDP):
                if not isinstance (self.src, tuple):
                    self.src = self.src, v["sport"]
                    self.dst = self.dst, v["dport"]
                v = v["pl"]
                continue
            elif isinstance (v, CoAP):
                self.coap = v
            elif isinstance (v, Ieee802154):
                self.src = v["src"]
                self.dst = v["dst"]
                v = v["pl"]
                continue
            elif isinstance (v, SixLowpan) or isinstance (v, LinuxCookedCapture) or isinstance (v, NullLoopback):
                try:
                    v = v["pl"]
                    continue
                except KeyError:
                    pass

            break


def get_implemented_testcases(testcase_id = None, no_verbose = None):
    """
    :return:
    -List of descriptions of test cases
    Each element of the list is composed of:
        -tc_identifier
        -tc_objective
        -tc_sourcecode
    """

    testcases,_= import_testcases(testcase_id)
    ret = []
    for tc in testcases:
        if no_verbose:
            ret.append((tc.__name__, tc.get_objective(), ''))
        else:
            ret.append((tc.__name__ ,tc.get_objective(), inspect.getsource(tc)))
    return ret


def import_testcases(testcase_id = None):
    """
    Assumptions:
    -test cases are defined inside a file, each file contains only one test case
    -names of the file and class must match
    -all test cases must be named TD_*

    :param testcase_id:

    Imports test cases classes from TESTCASES_DIR named TD*
    Returns:
        tuple of list:
        ( testcases , obsoletes )
    """

    # TODO take a list as a param and return corresponding testcases classes respecting the order.
    SEARCH_STRING = 'td*.py'
    tc_plugins = {}
    testcases = []
    obsoletes = []


    prv_wd = getcwd()
    chdir( prv_wd + TESTCASES_SUBDIR )

    #  find files named "TD*" or testcase_id (if provided) in TESTCASES_DIR
    dir_list = glob.glob(SEARCH_STRING)
    modname_test_list = [path.basename(f)[:-3] for f in dir_list if path.isfile(f)]
    modname_test_list.sort()

    if testcase_id:
        if testcase_id.lower() in modname_test_list:
            modname_test_list = [testcase_id]

            # filename not found in dir
        else:
            # move back to the previously dir
            chdir(prv_wd)
            raise FileNotFoundError("Testcase : " + testcase_id + " couldn't be found")

    # import sorted list
    for modname in modname_test_list:
        # note that the module is always lower case and the plugin (class) is upper case (ETSI naming convention)
        tc_plugins[modname] = getattr(import_module(modname.lower()), modname.upper())
        if tc_plugins[modname].obsolete:
            obsoletes.append (tc_plugins[modname])
        else:
            testcases.append (tc_plugins[modname])

    # move back to the previously dir
    chdir(prv_wd)

    assert all (isinstance (t, type) and issubclass (t,TestCase) for t in testcases)

    if obsoletes:
        sys.stderr.write ("%d obsolete testcases:\n" % len (obsoletes))
        for tc_type in obsoletes:
            sys.stderr.write ("\t%s\n" % tc_type.__name__)

    return testcases , obsoletes


def analyse_file (filename):
    testcases, _ = import_testcases()
    # read the frame
    # TODO: filter uninteresting frames ? (to decrease the load)
    with Data.disable_name_resolution():
        frames = Frame.create_list (PcapReader (filename))

        for f in frames:
            print ("%5d %s" % (f.id, f.msg.summary()))

        # malformed frames
        malformed = list (filter ((lambda f: f.exc), frames))

        print ("\n%d malformed frames" % len (malformed))
        for f in malformed:
            print ("%5d %s" % (f.id, f.exc))

        tracker = Tracker (frames)
        conversations = tracker.conversations
        ignored = tracker.ignored_frames
    #   sys.exit(1)
    #   conversations, ignored = extract_coap_conversations (frames)

        print ("\n%d ignored frames" % len (ignored))
        for f in ignored:
            print ("%5d %s" % (f.id, f.msg.summary()))

        print ("\n%d CoAP conversations" % len (conversations))
        for t in conversations:
            print ("    ---- Conversation %d    %s ----" % (t.id, t.tag))
            for f in t:
                print ("    %5d %s" % (f.id, f.msg.summary()))

        conversations_by_pair = proto_specific.group_conversations_by_pair (conversations)

        print ("\nTestcase results")
        results_by_pair = {}
        for pair, conversations in conversations_by_pair.items():
            pair_results = []
    #       print (pair, conversations)
            print ("---- Pair  %s -> %s ----" % pair)
            for tc_type in testcases:
                print (" --- Testcase %s ---" % tc_type.__name__)
                tc_results = []
                for tr in conversations:
                    tc = tc_type (tr)
                    if tc.verdict:
                        print ("    -- Conversation %d -> %s --" % (tc.conversation.id, tc.verdict))
                        for line in tc.text.split("\n"):
                            print ("\t" + line)
                        tc_results.append (tc)
                pair_results.append (tc_results)

    #   print (pair_results)

reg_frame = re.compile ("<Frame  *(\d+):")
reg_verdict = re.compile (r"    \[ *([a-z]+) *\]")


# TODO make unitest for analyse!
def analyse_file_rest_api(filename, urifilter = False, exceptions = None, regex = None, profile = "client", verbose=False):
    """
    :param filename:
    :param urifilter:
    :param exceptions:
    :param regex:
    :param profile:
    :param verbose: boolean, if true method returns verdict description (which may be very verbose)
    :return: tuple

    example:
    [('TD_COAP_CORE_03', 'fail', [21, 22]), 'verdict description']

    NOTES:
     - allows multiple ocurrences of the testcase, returns as verdict:
            - fail: if at least one on the occurrences failed
            - inconclusive : if all ocurrences returned a inconv verdict
            - pass: all occurrences are inconclusive or at least one is PASS and the rest is inconclusive
    """
    testcases, _ = import_testcases(regex)
    my_testcases = [t for t in testcases if t.reverse_proxy == (profile == "reverse-proxy") ]

    if regex is not None:
        try:
            re_regex = re.compile (regex, re.I)
        except Exception as e:
            return "Error: regular expression %r is invalid (%s)" % (regex, e)

    my_testcases = list (filter ((lambda t: re_regex.search(t.__name__)), my_testcases))

    if not my_testcases:
        return "regular expression %r did not yield any testcase" % regex
    force = len (my_testcases) == 1

    with Data.disable_name_resolution():
        frames = Frame.create_list (PcapReader (filename))

        # malformed frames
        malformed = list (filter ((lambda f: f.exc), frames))
        tracker = Tracker (frames)
        conversations = tracker.conversations
        ignored = tracker.ignored_frames


        results = []

        conversations_by_pair = proto_specific.group_conversations_by_pair(conversations)
        results_by_pair = {}
        for pair, conversations in conversations_by_pair.items():
            pair_results = []
            for tc_type in my_testcases:
                tc_results = []

                # we run the testcase for each conversation, meaning that one type of TC can have more than one result!
                for tr in conversations:
                    tc = tc_type(tr, urifilter, force)
                    if tc.verdict:

                        tc_results.append(tc)

                        # remember the exception
                        if hasattr(tc, "exception") and exceptions is not None:
                            exceptions.append(tc)

                pair_results.append(tc_results)

        verdicts = None, "inconclusive", "pass", "fail", "error"
        for tc_type, tc_results in filter(lambda x: regex in x[0].__name__, zip(my_testcases, pair_results)):
            review_frames=[]
            v = 0
            for tc in tc_results:
                # all the failed frames for a TC, even if they are from different conversations!
                review_frames = tc.failed_frames

                new_v = verdicts.index(tc.verdict)
                if new_v > v:
                    v = new_v

            v_txt = verdicts[v]
            if v_txt is None:
                v_txt = "none"

            if verbose:
                results.append( (type(tc).__name__, v_txt, list(review_frames),tc.text))
            else:
                results.append((type(tc).__name__, v_txt, list(review_frames),''))
            # TODO clean list(review_frames)  tc.review_frames_log , tc.review_frames_log in module proto_specific

        return results


def analyse_file_html (filename, orig_name, urifilter = False, exceptions = None, regex = None, profile = "client"):
    """
    TODO
    Args:
        filename:
        orig_name:
        urifilter:
        exceptions:
        regex:
        profile:

    Returns:

    """
    testcases, _ = import_testcases()

    logger = HTMLLogger()

    with XHTML10Generator () as g:

        def log_text (text):
            if not isinstance (text, str):
                text = str (text)
            for line in text.split("\n"):
                mo = reg_frame.match (line)
                if mo:
                    g.a(href="#frame"+mo.group(1))(line + "\n")
                    continue
                mo = reg_verdict.match (line)
                if mo:
                    g.span (**{"class": mo.group(1)})(line)
                    g ("\n")
                elif line.endswith ("Mismatch"):
                    g.span (**{"class": "mismatch"})(line)
                    g ("\n")
                elif line.startswith ("Chaining to conversation "):
                    g ("\n")
                    g.span (**{"class": "chaining"})(line)
                    g ("\n")
                else:
                    g (line)
                    g ("\n")

        with g.head:
            g.title ("CoAP interoperability test results")
            g.meta (**{"http-equiv": "Content-Type", "content": "text/html; charset=utf-8"})

            g.style ("""
a {color: inherit; text-decoration: inherit}
.pass {color: green;}
.inconclusive {color: #e87500;}
.fail {color: red;}
.error {color: red;}
.mismatch {color: #803000;}
.chaining {color: #808080; font-style: italic;}
.bgpass {background-color: #B0FFB0;}
.bginconclusive {background-color: #FFB080;}
.bgfail {background-color: #FF9090;}
.bgerror {background-color: #FF9090;}
.bgnone {background-color: #FFFFB0;}
td {padding-left:15px; padding-right:15px; padding-top:3px; padding-bottom:3px;}
th {padding-left:15px; padding-right:15px; padding-top:10px; padding-bottom:10px;}
table {border: 1px solid; border-spacing: 0px; }
""", type = "text/css")

        g.h1 ("CoAP interoperability test results")

        g.pre("""Tool version:   %s
File:           %s
Date:           %s
URI filter:     %s
Regex:      %r
""" % ( TOOL_VERSION,
    (orig_name if orig_name else "(unknown)"),
    time.strftime ("%a, %d %b %Y %T %z"),
    ("enabled" if urifilter else "disabled"),
    regex),

    style = "line-height: 150%;")

        my_testcases = [t for t in testcases if t.reverse_proxy == (profile == "reverse-proxy") ]

        if regex is not None:
            try:
                re_regex = re.compile (regex, re.I)
            except Exception as e:
                g.b("Error: ")
                g("regular expression %r is invalid (%s)" % (regex, e))
                return

            my_testcases = list (filter ((lambda t: re_regex.search(t.__name__)), my_testcases))

            if not my_testcases:
                g.b("Warning: ")
                g("regular expression %r did not yield any testcase" % regex)


        force = len (my_testcases) == 1

        g.h2 ("Summary")
        with g.table(id="summary", border=1): # FIXME: bug xmlgen
            pass

        with Data.disable_name_resolution():
            frames = Frame.create_list (PcapReader (filename))

            g.h2("File content (%d frames)" % len (frames))
            with g.pre:
                for f in frames:
                    log_text (f)

            # malformed frames
            malformed = list (filter ((lambda f: f.exc), frames))

            g.h2("Malformed frames (%d)" % len (malformed))
            with g.pre:
                for f in malformed:
                    log_text (f)
                    g(" %s\n" % f.exc)

            tracker = Tracker (frames)
            conversations = tracker.conversations
            ignored = tracker.ignored_frames
        #   sys.exit(1)
        #   conversations, ignored = extract_coap_conversations (frames)

            g.h2 ("Ignored frames (%d)" % len (ignored))
            with g.pre:
                for f in ignored:
                    log_text (f)

            g.h2 ("CoAP conversations (%d)" % len (conversations))
            for t in conversations:
                g.h3 ("Conversation %d %s" % (t.id, t.tag))
                with g.pre:
                    for f in t:
                        log_text (f)

            conversations_by_pair = proto_specific.group_conversations_by_pair (conversations)

            g.h2 ("Testcase results")

            with g.script (type="text/javascript"):
                g('''
var t=document.getElementById("summary");
var r;
var c;
var a;
''')
            results_by_pair = {}
            for pair, conversations in conversations_by_pair.items():
                pair_results = []
                pair_txt = "%s vs %s" % tuple (map (Resolver.format, pair))
                g.h3 (pair_txt)
                for tc_type in my_testcases:
                    tc_results = []
                    g.a(name="%x" % id(tc_results))
                    g.h4 ("Testcase %s  -  %s" % (tc_type.__name__, tc_type.get_objective()))
                    for tr in conversations:
                        tc = tc_type (tr, urifilter, force)
                        if tc.verdict:
                            with g.h5:
                                g ("Conversation %d -> "% tc.conversation.id)
                                g.span (tc.verdict, **{"class": tc.verdict})
                            with g.pre:
                                log_text (tc.text)

                            tc_results.append (tc)

                            # remember the exception
                            if hasattr (tc, "exception") and exceptions is not None:
                                exceptions.append (tc)

                    pair_results.append (tc_results)

                with g.script (type="text/javascript"):

                    g('''
r=t.insertRow(-1);
c=document.createElement ("th");
c.innerHTML=%s;
c.colSpan=4;
r.appendChild(c);
''' % repr (pair_txt))
                    verdicts = None, "inconclusive", "pass", "fail", "error"
                    for title, func in (
                            ("ETSI interoperability test scenarios",(lambda x: "IRISA" not in x[0].__name__)),
                            ("IRISA interoperability test scenarios",(lambda x: "IRISA" in x[0].__name__))
                        ):
                        g('''
r=t.insertRow(-1);
c=document.createElement ("th");
c.innerHTML=%s;
c.colSpan=4;
r.appendChild(c);
''' % repr(title))
                        for tc_type, tc_results in filter(func,zip(my_testcases,pair_results)):
                            v = 0
                            for tc in tc_results:
                                new_v = verdicts.index (tc.verdict)
                                if new_v > v:
                                    v = new_v
                            v_txt = verdicts[v]
                            if v_txt is None:
                                v_txt = "none"
#TODO: factorise that w/ a function
                            g('''
r=t.insertRow(-1);

a=document.createElement ("a")
a.href=("#%x")
a.innerHTML=%s
r.insertCell(-1).appendChild(a);

a=document.createElement ("a")
a.href=("#%x")
a.innerHTML=%s
r.insertCell(-1).appendChild(a);

a=document.createElement ("a")
a.href=("#%x")
a.innerHTML=%s
r.insertCell(-1).appendChild(a);

a=document.createElement ("a")
a.href=("#%x")
a.innerHTML=%s
c=r.insertCell(-1)
c.appendChild(a);
c.className=%s;
''' % (
                                id (tc_results),
                                repr (tc_type.__name__),
                                id (tc_results),
                                repr (tc_type.get_objective()),
                                id (tc_results),
                                repr ("%d occurence(s)" % len (tc_results)),
                                id (tc_results),
                                repr(v_txt),
                                repr("bg" + v_txt),
                            ))


                    """
                    pair_results = []
                    g.h3("%s vs %s" % pair)
                    for tc_type in my_testcases:
                        g.h4 ("Testcase %s" % tc_type.__name__)
                        tc_results = []
                        for tr in conversations:
                            tc = tc_type (tr)
                            if tc.verdict:
                                with g.h5:
                                    g ("Conversation %d -> "% tc.conversation.id)
                                    g.span (tc.verdict, **{"class": tc.verdict})
                                with g.pre:
                                    log_text (tc.text)

                                tc_results.append (tc)

                        pair_results.append (tc_results)
                """

            g.h2 ("Frames details")

            for f in frames:
                with g.pre:
                    g.a(name = "frame%d" % f.id)
                    g.b()("\n%s\n\n" % f)

                    b = f.msg.get_binary()
                    for offset in range(0, len(b), 16):
                        values = ["%02x" % v for v in b[offset:offset+16]]
                        if len(values) > 8:
                            values.insert (8, " ")

                        g("         %04x     %s\n" % (offset, " ".join(values)))
                logger.display_value (g, f.msg.get_value())
                g(" ")
                g.br()
                g.hr()


def basic_dissect_pcap_to_list (filename, protocol_selection=None):
    """
    :param filename: filename of the pcap file to be dissected
    :param protocol_selection:  protocol class for filtering purposes (inheriting from packet Value)
    :return: list of tuples with basic info about frames:
    [
        (13, '[127.0.0.1                        -> 127.0.0.1                       ] CoAP [CON 38515] GET /test'),
        (14, '[127.0.0.1                        -> 127.0.0.1                       ] CoAP [ACK 38515] 2.05 Content '),
        (21, '[127.0.0.1                        -> 127.0.0.1                       ] CoAP [CON 38516] PUT /test'),
        (22, '[127.0.0.1                        -> 127.0.0.1                       ] CoAP [ACK 38516] 2.04 Changed ')]
    ]

    """
    # read the frame
    # TODO: filter uninteresting frames ? (to decrease the load)
    with Data.disable_name_resolution():
        frames = Frame.create_list (PcapReader (filename))
        response=[]

        # content of the response, TODO make this generic for any protocol
        if protocol_selection and protocol_selection is CoAP:
            selected_frames = [f for f in frames if f.coap]
        else:
            selected_frames = frames

        for f in selected_frames:
                response.append ((f.id , f.msg.summary()))
        # malformed frames
        # malformed = list (filter ((lambda f: f.exc), frames))
    return response


def value_to_list(l: list, value: Value, extra_data=None, layer_dict:dict=None, is_option=False):

    # points to packet
    if isinstance(value, PacketValue):

        od = OrderedDict()

        if is_option:
            od['Option'] = value.get_variant().__name__
        else:
            od['_type'] = 'protocol'
            od['_protocol'] = value.get_variant().__name__

        l.append(od)

        i = 0
        for f in value.get_variant().fields():
            value_to_list(l, value[i], f.name, od)
            i += 1

    # TODO test this
    elif isinstance(value, ListValue):
        prot_options = []
        for i in range(0, len(value)):
            value_to_list(prot_options, value[i], is_option=True)
        layer_dict['Options'] = prot_options

    # it's a field
    else:
        layer_dict[extra_data] = str(value)


def dissect_pcap_to_list(filename, protocol_selection=None):
    """
    :param filename: filename of the pcap file to be dissected
    :param protocol_selection:  protocol class for filtering purposes (inheriting from packet Value)
    :return: List of frames (frames as Ordered Dicts)
    """

    if protocol_selection:
        assert issubclass(protocol_selection, PacketValue)

    frame_list = []

    # for speeding up the process
    with Data.disable_name_resolution():

        frames = Frame.create_list(PcapReader(filename))

        for f in frames:

            # if we are filtering and frame doesnt contain protocol
            # then skip frame
            # TODO make this generic for any type of protocol
            if protocol_selection and not f.coap:
                pass
            else:
                frame = OrderedDict()
                frame['_type'] = 'frame'
                frame['id'] = f.id
                frame['timestamp'] = f.ts
                frame['error'] = f.exc
                frame['protocol_stack'] = []
                value_to_list(frame['protocol_stack'], f.msg.get_value())
                frame_list.append(frame)
    return frame_list


################################################################################
# Main()
################################################################################

if __name__ == "__main__":

    import pprint,json
    PCAP_test2 = getcwd()+ "/tests/test_dumps/coap/TD_COAP_CORE_01_PASS.pcap"
    PCAP_test3 = getcwd() + "/tests/test_dumps/coap/coap_get_migled_with_tcp_traffic.pcap"



    #analyse_file(PCAP_test3)

    #get test case implementation
    #a= get_implemented_testcases('td_coap_core_01')
    #for f in a:
    #    print(a[0][2])

    #print(analyse_file_rest_api(PCAP_test2,False,None,"TD_COAP_CORE_17","client"))

    #print(dissect_pcap_to_list(PCAP_test3, None))
    #print(type(basic_dissect_pcap_to_list(PCAP_test3, None)))
    #print(basic_dissect_pcap_to_list(PCAP_test3, CoAP))

    print('************************')
    print('**API** ')
    print('************************')
    #print(basic_dissect_pcap_to_list(PCAP_test3, CoAP))
    print(analyse_file_rest_api(PCAP_test3, False, None, "TD_COAP_CORE_03", "client",True))
    #print(analyse_file_rest_api(PCAP_test3, False, None, "", "client",False))
    #print(analyse_file_rest_api(PCAP_error,None,None,"TD_COAP_CORE_01","client"))

    #a= get_implemented_testcases('td_coap_coreasd_01')
    #for f in a:
    #    print(a)
